import pydicom 

class RTROIObservationsSequence :
    """a class to represent RTROIObservationsSequence
    """

    def __init__(self, results:dict, number_of_roi:int):
        """constructor

        Args:
            results (dict): [dict generated by 'generate_dict.py' ]
            number_of_roi (int): [total number of ROI]
        """
        self.number_of_roi = number_of_roi 
        self.results = results


    def create_RTROIObservationsSequence(self) -> pydicom.Sequence:
        """method to generate RTROIObservationsSequence

        Returns:
            [pydicom.Sequence]: [return RTTOIObservationsSequence ]
        """
        RTROIObservationsSequence = pydicom.sequence.Sequence()
        for number_roi in range(1, self.number_of_roi +1) : 
            dataset = pydicom.dataset.Dataset()
            dataset.ReferencedROINumber = number_roi
            dataset.ObservationNumber = number_roi 

            dataset.RTROIInterpretedType = self.results["segmentAttributes"][0][number_roi-1]['RTROIInterpretedType']
            dataset.ROIInterpreter = ''
            
            RTROIObservationsSequence.append(dataset)

        return RTROIObservationsSequence 


from dicom_to_cnn.model.reader.Series import Series
from dicom_to_cnn.model.reader.Instance import Instance 
import os 
from math import exp, log, pow
from datetime import datetime, timedelta
import numpy as np 
import SimpleITK as sitk 

class SeriesPT(Series):
    """Specific methods to read PET series
    Store radiopharmaceutical data
    Check Attenuation conrrection
    SUV / SUL calculation
    Get Series as 32bits Nifti SUV/SUL
    

    Arguments:
        Series {[class]} -- [description]
    """

    def __init__(self, path:str):
        """Constructor

        Arguments:
            path {Str} -- Path folder of series
        """
        super().__init__(path)
        

    def get_minimum_acquisition_datetime(self) -> datetime:
        """Get earlier acquisition datetime of the PET serie

        Returns : 
            [datetime] -- [Return the earlier acquisition datetime, or "Undefined" if no acquisition_time or no acquisition date]
        """
        liste = []
        try : 
            for filename in self.file_names : 
                instanceData = Instance(os.path.join(self.path,filename), load_image=True)
                datetime = instanceData.get_acquisition_date() + instanceData.get_acquisition_time()
                parsed_datetime = self.__parse_datetime(datetime)
                liste.append(parsed_datetime)

            return min(liste)

        except Exception : 
            return "Undefined"
        


    def get_series_details(self) -> dict:
        """Add Pharmaceuticals data to common series details

        Returns:
            [dict] -- [Return the details of a SeriePT from the first Dicom]
        """
        details = super().get_series_details()
        dicomInstance = self.get_first_instance_metadata()
        self.radiopharmaceutical_details = {}
        self.radiopharmaceutical_details = dicomInstance.get_radiopharmaceuticals_tags()
        details['radiopharmaceutical'] = self.radiopharmaceutical_details
        self.pet_correction_details = {}
        self.pet_correction_details = dicomInstance.get_pet_correction_tags()
        details['pet_correction'] = self.pet_correction_details
        self.philips_tag = {}
        self.philips_tag = dicomInstance.get_philips_private_tags()
        details['philips_tags'] = self.philips_tag

        return details
        

    @classmethod
    def __parse_datetime(cls, date_time:str) -> datetime:
        """class method to parse date time 

        Args:
            date_time (str): [date and time in  "%Y%m%d%H%M%S" format ]

        Returns:
            [datetime]: [return parsed datetime]]
        """
        #remove microsecond at it is inconstant over dicom
        if '.' in date_time : 
            date_time = date_time[0 : date_time.index('.')]
        #parse datetime to date objet
        return datetime.strptime(date_time, "%Y%m%d%H%M%S")



    def __calculateSUVFactor(self) -> float:
        """
        Calcul of  SUV factor

        Returns:
            [float] -- [return SUV factor or "Calcul SUV impossible" if there is "Undefined" value in tags]
        """
        
        series_details = self.get_series_details()
        units = series_details['series']['Units']
        known_units = ['GML', 'BQML', 'CNTS']
        if units not in known_units : 
            raise Exception ('Unknown PET Units')

        if units == 'GML' : return 1
        elif units == 'CNTS' :
            philips_suv_bqml = series_details['philips_tags']['PhilipsBqMlFactor']
            philips_suv_factor = series_details['philips_tags']['PhilipsSUVFactor']
            if (philips_suv_factor != 'Undefined') : return philips_suv_factor
            if (philips_suv_factor == 'Undefined' and philips_suv_bqml == 'Undefined') : raise Exception('Missing Philips private Factors')
 
        if series_details['study']['PatientWeight'] == None : 
            raise Exception('Missing patient weight')
        patient_weight = series_details['study']['PatientWeight'] * 1000 #kg to g conversion
        series_time = series_details['series']['SeriesTime']
        series_date = series_details['series']['SeriesDate']
        series_datetime = series_date + series_time 

        series_datetime = self.__parse_datetime(series_datetime)
        acquisition_datetime = self.get_minimum_acquisition_datetime()
        acquisition_date = series_details['series']['AcquisitionDate']
        decay_correction = series_details['series']['DecayCorrection']
        radionuclide_half_life = series_details['radiopharmaceutical']['RadionuclideHalfLife']
        total_dose = series_details['radiopharmaceutical']['TotalDose']

        radiopharmaceutical_start_date_time = series_details['radiopharmaceutical']['RadiopharmaceuticalStartDateTime']

        
        if radiopharmaceutical_start_date_time == 'Undefined' or radiopharmaceutical_start_date_time == '' : 
            #If startDateTime not available use the deprecated statTime assuming the injection is same day than acquisition date
            radiopharmaceutical_start_time = series_details['radiopharmaceutical']['RadiopharmaceuticalStartTime']
            radiopharmaceutical_start_date_time = acquisition_date + radiopharmaceutical_start_time 
            
        
        radiopharmaceutical_start_date_time = self.__parse_datetime(radiopharmaceutical_start_date_time)
        
        if (total_dose == 'Undefined' or total_dose == None or  acquisition_datetime== 'Undefined' 
            or patient_weight == 'Undefined' or patient_weight == None or radionuclide_half_life == 'Undefined' ) :
            raise Exception('Missing Radiopharmaceutical data or patient weight')
        
        #Determine Time reference of image acqusition 
        acquisition_hour = series_datetime
        if (acquisition_datetime != 'Undefined'
             and (acquisition_datetime - series_datetime).total_seconds() < 0 and units == 'BQML') : 
            acquisition_hour = acquisition_datetime

        
        #Calculate decay correction
        if decay_correction == 'START' : 
            delta = (acquisition_hour - radiopharmaceutical_start_date_time)
            delta = delta.total_seconds()
            if (delta < 0) : raise("Acqusition time before injection time")
            decay_factor = exp(-delta * log(2) / radionuclide_half_life)

        #If decay corrected from administration time no decay correction to apply
        elif decay_correction == 'ADMIN' : 
            decay_factor = 1


        else : raise Exception('Unknown Decay Correction methode')
        
        suv_conversion_factor = 1/((total_dose * decay_factor) / patient_weight)
        
        if units == 'CNTS' : return philips_suv_bqml * suv_conversion_factor
        else : return suv_conversion_factor


    def calculateSULFactor(self) -> float:
        """Calcul SUL Factor

        Returns:
            [float] -- [Return SUL Factor or "Calcul SUL impossible" if there is "Undefined" value in tags]
        """
        series_details = self.get_series_details()
        patient_sex = series_details['patient']['PatientSex']
        patient_weight = series_details['study']['PatientWeight']
        patient_height = series_details['study']['PatientHeight']
        if (patient_sex == 'Undefined' or patient_height == 'Undefined' or patient_weight == 'Undefined' or patient_height == 0.0) : 
            raise Exception('Missing Height or Weight to calculate SUL')
        bmi =  patient_weight / pow(patient_height, 2)
        if patient_sex == 'F' :
            return 9270 / (8780 + 244 * bmi)
        elif patient_sex == 'M':
            return 9270 / (6680 + 216 * bmi)
        else :
            raise Exception('Unknown Sex String')
    

    def is_corrected_attenuation(self) -> bool :
        """If PET Series is attenuation corrected

        Returns:
            [bool] -- [True if attenuation corrected]
        """
        series_details = self.get_series_details()
        corrected_image = series_details['pet_correction']
        if 'ATTN' in corrected_image : return True
        else : return False


    
    def get_numpy_array(self) -> np.ndarray:
        """method to get 3D numpy array 

        Returns:
            [ndarray] -- [return array of the SeriesPT with SUV and SUL factor in 32bis npArray ]
        """
        numpy_array = super().get_numpy_array()
        return numpy_array 


    def set_ExportType(self, export_type:str = 'suv'):
        """set the export value for PET array

        Args:
            export_value (str, optional): ['raw', 'suv', 'sul']. Defaults to 'suv'.
        """
        self.export_type = export_type


    def export_nifti(self, file_path:str):
        """method to export/save ndarray of series to nifti format

        Args:
            file_path (str): [directory+filename of the nifti]
        """
        pet_array = super().get_numpy_array()

        if self.export_type == 'raw' : 
            pass

        elif self.export_type == 'suv' : 
            try : 
                pet_array = pet_array * self.__calculateSUVFactor() 
            except Exception as err : 
                print("Error generating result array (suv mode)", err)

        elif self.export_type == 'sul' : 
            try : 
                pet_array = pet_array * self.__calculateSUVFactor() * self.calculateSULFactor() 
            except Exception as err :
                print("Error generating result array (sul mode)", err) 


        sitk_img = sitk.GetImageFromArray( np.transpose(pet_array, (2,0,1) ))  
        sitk_img = sitk.Cast(sitk_img, sitk.sitkFloat32)  
        original_pixel_spacing = self.instance_array[0].get_pixel_spacing()   
        original_direction = self.instance_array[0].get_image_orientation()
        sitk_img.SetDirection( (float(original_direction[0]), float(original_direction[1]), float(original_direction[2]), 
                                    float(original_direction[3]), float(original_direction[4]), float(original_direction[5]), 
                                    0.0, 0.0, 1.0) )
        sitk_img.SetOrigin( self.instance_array[0].get_image_position() )
        sitk_img.SetSpacing( (original_pixel_spacing[0], original_pixel_spacing[1], self.get_z_spacing()) )
        sitk.WriteImage(sitk_img, file_path)





import os 
from dicom_to_cnn.enums.TagEnum import *
from dicom_to_cnn.enums.SopClassUID import *
from dicom_to_cnn.model.reader.Series import Series 
from dicom_to_cnn.model.reader.SeriesCT import SeriesCT
from dicom_to_cnn.model.reader.SeriesPT import SeriesPT 
from dicom_to_cnn.model.reader.Instance import Instance  

def get_series_object(path:str):
    """
    class method to generate a Series object

    Args:
        path (str): [path folder of series]

    Returns:
        [Series]: [return a Series object]
    """
    try:
        first_file_name = os.listdir(path)[0]
    except Exception as err:
        print(err)
        print(path)
    first_instance = Instance( os.path.join(path,first_file_name) )
    sop_class_uid = first_instance.get_sop_class_uid()
    if(sop_class_uid == ImageModalitiesSOPClass.PT.value or sop_class_uid == ImageModalitiesSOPClass.EnhancedPT.value):
        return SeriesPT(path)
    elif (sop_class_uid == ImageModalitiesSOPClass.CT.value or sop_class_uid == ImageModalitiesSOPClass.EnhancedCT.value):
        return SeriesCT(path)
    else : return Series(path)



from dicom_to_cnn.model.petctviewer.RoiPolygon import RoiPolygon
from dicom_to_cnn.model.petctviewer.RoiElipse import RoiElipse
from dicom_to_cnn.model.petctviewer.RoiNifti import RoiNifti
from dicom_to_cnn.model.petctviewer.Roi import Roi 


class RoiFactory():
    """Create ROI object from a CSV Roi details

    Returns:
        [RoiFactory] -- Factory for ROI instanciation according to csv details
    """

    def __init__(self, details:dict, volume_dimension:tuple, roi_number:int):
       """constructor

       Arguments:
           details {dict} -- Parsed ROI informations
           volume_dimension {tuple} -- Matrice dimension for Mask (should match DICOM input size, (shape x, shape y, shape z))
           roi_number {int} -- Integer number that will have to be written in the mask
       """
       self.details = details
       self.volume_dimension = volume_dimension
       self.roi_number = roi_number
    
    def read_roi(self) -> Roi:
        """Instanciate the Correct ROI type according to data

        Returns:
            [Roi] -- Roi object
        """
        first_slice = self.details['first_slice']
        last_slice = self.details['last_slice']
        point_list  = self.details['point_list']
        roi_number = self.roi_number
        type_number = self.details['type_number']
        volume_dimension = self.volume_dimension

        if (self.details['type_number'] == 1):
            return RoiPolygon( 1, first_slice, last_slice, roi_number, type_number, point_list, volume_dimension )

        elif (self.details['type_number'] == 11):
            return RoiElipse( 1, first_slice, last_slice, roi_number, type_number, point_list, volume_dimension )

        elif (self.details['type_number'] == 2):
            return RoiPolygon( 2, first_slice, last_slice, roi_number, type_number, point_list, volume_dimension )

        elif (self.details['type_number'] == 12):
            return RoiElipse( 2, first_slice, last_slice, roi_number, type_number, point_list, volume_dimension )

        elif (self.details['type_number'] == 3):
            return RoiPolygon( 3, first_slice, last_slice, roi_number, type_number, point_list, volume_dimension )

        elif (self.details['type_number'] == 13):
            return RoiElipse( 3, first_slice, last_slice, roi_number, type_number, point_list, volume_dimension )
        
        else:
            return RoiNifti( roi_number, point_list, volume_dimension )


    


import pydicom
import numpy as np
from dicom_to_cnn.enums.TagEnum import *
from dicom_to_cnn.enums.SopClassUID import *

class Instance:
    """A class to represent a Dicom file 
    """

    def __init__(self, path:str, load_image:bool=True):
        """Construct a Dicom file object

        Args:
            path (str): [path of instance dicom]
            load_image (bool, optional): [choose to load only metadata or metadata+image]. Defaults to True.
        """
        self.path = path
        if (load_image) : self.__load_full_instance()
        else : self.__load_metadata()

    def __load_metadata(self):
        """load only metadata 
        """
        self.dicomData = pydicom.dcmread(self.path, stop_before_pixels=True, force=True)
    
    def __load_full_instance(self):
        """load metadata and image 
        """
        self.dicomData = pydicom.dcmread(self.path, force=True)
  
    def get_series_tags(self) -> dict :
        """method to gather series tags 

        Returns:
            [dict]: [dictionnary of every series tags and value]
        """
        series_tags={}
        for tag_address in TagsSeries:
            if tag_address.value in self.dicomData :
                series_tags[tag_address.name] = self.dicomData[tag_address.value].value
            else : series_tags[tag_address.name] = "Undefined"
        series_tags[ImageType.ImageType.name] = self.get_image_type()
        series_tags[PixelSpacing.PixelSpacing.name] = self.get_pixel_spacing()
        return series_tags

    def get_patients_tags(self) -> dict :
        """method to gather patient tags 

        Returns:
            [dict]: [dictionnary of every patient tags and value]
        """
        patient_tags={}
        for tag_address in TagsPatient:
            if tag_address.value in self.dicomData : 
                patient_tags[tag_address.name] = self.dicomData[tag_address.value].value
                if( type(self.dicomData[tag_address.value].value) != str):
                    #patient name return PersonName3 object, convert it to string
                    patient_tags[tag_address.name] = str(patient_tags[tag_address.name])
            else : patient_tags[tag_address.name] = "Undefined"
        return patient_tags

    def get_studies_tags(self) -> dict :
        """method to gather study tags 

        Returns:
            [dict]: [dictionnary of every study tags and value]
        """
        studies_tags={}
        for tag_address in TagsStudy:
            if tag_address.value in self.dicomData : studies_tags[tag_address.name] = self.dicomData[tag_address.value].value
            else : studies_tags[tag_address.name] = "Undefined"
        return studies_tags

    def get_instance_tags(self) -> dict :
        """method to gather instance tags 

        Returns:
            [dict]: [dictionnary of every instance tags and value]
        """
        instance_tags={}
        instance_tags['PixelSpacing'] = self.get_pixel_spacing()
        instance_tags['ImagePosition'] = self.get_image_position()
        instance_tags['ImageOrientation'] = self.get_image_orientation()
        instance_tags['RescaleSlope'] = self.get_rescale_slope()
        instance_tags['RescaleIntercept'] = self.get_rescale_intercept()
        instance_tags['SOPInstanceUID'] = self.get_SOPInstanceUID()
        return instance_tags

    def get_radiopharmaceuticals_tags(self) -> dict :
        """method to gather radiopharmaceutical tags 

        Returns:
            [dict]: [dictionnary of every radiopharmaceutical tags and value]
        """
        radiopharmaceuticals_tags={}
        radiopharmaceutical_sequence = []
        try :
            radiopharmaceutical_sequence = self.dicomData[0x00540016][0]
        except Exception: 
            print("no Radiopharmaceuticals tags")
        for tag_address in TagsRadioPharmaceuticals:
            if tag_address.value in radiopharmaceutical_sequence : radiopharmaceuticals_tags[tag_address.name] = radiopharmaceutical_sequence[tag_address.value].value
            else : radiopharmaceuticals_tags[tag_address.name] = "Undefined"

        return radiopharmaceuticals_tags
    
    def get_philips_private_tags(self) -> dict :
        """method to gather philips private tags 

        Returns:
            [dict]: [dictionnary of every philips private tags and value]
        """
        philips_tags={}
        for tag_address in PhilipsPrivateTags:
            if tag_address.value in self.dicomData : 
                philips_tags[tag_address.name] = float(self.dicomData[tag_address.value].value)
            else : philips_tags[tag_address.name] = "Undefined"
        return philips_tags

    def get_pet_correction_tags(self) -> dict :
        """method to gather pet correction tags 

        Returns:
            [dict]: [dictionnary of every pet correction tags and value]
        """
        if TagPTCorrection.CorrectedImage.value in self.dicomData :
            return list(self.dicomData[TagPTCorrection.CorrectedImage.value].value)
        else: return "Undefined"

    def get_sop_class_uid(self) -> str :
        """method to get sop class uid 

        Raises:
            Exception: [raise exception if undefined SOPClassUID]

        Returns:
            [str]: [return SOPClassUID value if defined]
        """
        if 'SOPClassUID' in self.dicomData.dir() : return self.dicomData.SOPClassUID
        else : raise Exception('Undefined SOP Class UID')

    def is_secondary_capture(self) -> bool :
        """check if SOPClassUID in CapturesSOPClass list

        Returns:
            [bool]: [return True if SOPCLassUID in CapturesSOPClass list, False instead]
        """
        return True if self.get_sop_class_uid in CapturesSOPClass else False

    def get_SOPInstanceUID(self) -> str:
        """get SOPInstanceUID value

        Returns:
            str: [description]
        """
        return self.dicomData[TagsInstance.SOPInstanceUID.value].value

    def get_rescale_slope(self) -> str:
        """get rescale slope value

        Returns:
            str: [description]
        """
        return self.dicomData[TagsInstance.RescaleSlope.value].value

    def get_rescale_intercept(self) -> str:
        """get rescale intercept value

        Returns:
            str: [description]
        """
        return self.dicomData[TagsInstance.RescaleIntercept.value].value

    def get_image_orientation(self) -> list:
        """get image orientation value x,y,z

        Returns:
            list: [description]
        """
        return list(self.dicomData[TagsInstance.ImageOrientation.value].value)
    
    def get_image_position(self) -> list:
        """get image position value x,y,z

        Returns:
            list: [description]
        """
        return list(self.dicomData[TagsInstance.ImagePosition.value].value)

    def get_pixel_spacing(self) -> list:
        """get pixel spacing value x,y

        Returns:
            list: [description]
        """
        return list(self.dicomData[PixelSpacing.PixelSpacing.value].value)    
        
    def get_image_type(self) -> list:
        """get Image type value 

        Returns:
            list: [description]
        """
        return list(self.dicomData[ImageType.ImageType.value].value)

    def get_series_instance_uid(self) -> str:
        """get Series Instance UID value

        Returns:
            str: [description]
        """
        return self.dicomData[TagsSeries['SeriesInstanceUID'].value].value
    
    def get_accession_number(self) -> str:
        """get Accession Number value

        Returns:
            str: [description]
        """
        return self.dicomData[TagsStudy['AccessionNumber'].value].value 

    def get_patient_name(self) -> str:
        """get Patient Name value

        Returns:
            str: [description]
        """
        return self.dicomData[TagsPatient['PatientName'].value].value

    def get_patient_id(self) -> str:
        """get patient ID value

        Returns:
            str: [description]
        """
        return self.dicomData[TagsPatient['PatientID'].value].value

    def get_patient_birth_date(self) -> str:
        """get Patient Birth Date value

        Returns:
            str: [description]
        """
        return self.dicomData[TagsPatient['PatientBirthDate'].value].value

    def get_patient_sex(self) -> str:
        """get Patient sex value

        Returns:
            str: [description]
        """
        return self.dicomData[TagsPatient['PatientSex'].value].value 

    def get_study_date(self) -> str:
        """get study date value

        Returns:
            str: [description]
        """
        return self.dicomData[TagsStudy['StudyDate'].value].value

    def get_study_description(self) -> str:
        """get Study Description value

        Returns:
            str: [description]
        """
        try : 
            return self.dicomData[TagsStudy['StudyDescription'].value].value 
        except Exception : 
            return ''

    def get_study_id(self) -> str:
        """get study id value

        Returns:
            str: [description]
        """
        return self.dicomData[TagsStudy['StudyID'].value].value 

    def get_study_instance_uid(self) -> str:
        """get study instance uid value

        Returns:
            str: [description]
        """
        return self.dicomData[TagsStudy['StudyInstanceUID'].value].value

    def get_study_time(self) -> str:
        """ get study time value 

        Returns:
            str: [description]
        """
        return self.dicomData[TagsStudy['StudyTime'].value].value

    def get_number_rows(self) -> str:
        """get number of rows value

        Returns:
            str: [description]
        """
        return self.dicomData.Rows

    def get_number_columns(self) -> str:
        """get number of columns value

        Returns:
            str: [description]
        """
        return self.dicomData.Columns

    def get_frame_of_reference_uid(self) -> str:
        """get frame of reference uid value

        Returns:
            str: [description]
        """
        return self.dicomData.FrameOfReferenceUID

    def get_acquisition_date(self) -> str:
        """get acquisition date value

        Returns:
            str: [description]
        """
        if "AcquisitionDate" in self.dicomData : 
            return self.dicomData.AcquisitionDate
        else : return "Undefined"
        
    def get_acquisition_time(self) -> str:
        """get acquisition time value

        Returns:
            str: [description]
        """
        if 'AcquisitionTime' in self.dicomData : 
            return self.dicomData.AcquisitionTime
        else : return "Undefined"

    def get_referring_physician_name(self) -> str:
        """get referring physician name value

        Returns:
            str: [description]
        """
        if "ReferringPhysicianName" in self.dicomData : return self.dicomData.ReferringPhysicianName
        else : return ''

    def get_specific_character_set(self) -> str:
        """get specific chracter set value

        Returns:
            str: [description]
        """
        if "SpecificCharacterSet" in self.dicomData : return self.dicomData.SpecificCharacterSet
        else : return ''

    def get_physicians_of_record(self) -> str:
        """geet physicians of record value

        Returns:
            str: [description]
        """
        if "PhysiciansOfRecord" in self.dicomData : return self.dicomData.PhysiciansOfRecord
        else : return ''
 
    def is_image_modality(self) -> bool :
        """check if SOPClassUID in sop values list 

        Returns:
            [bool]: [return True if SOPCLassUID in sop values list]
        """
        sop_values = set(item.value for item in ImageModalitiesSOPClass)
        return True if self.get_sop_class_uid() in sop_values else False

    def get_image_nparray(self) -> np.ndarray:
        """get instance image ndarray 

        Raises:
            Exception: [raise Exception if SOPClassUID not in sop values list]

        Returns:
            [ndarray]: [return instance image ndarray]
        """
        if self.is_image_modality() == False : 
            raise Exception('Not Image Modality')
        else:
            pixel_array = self.dicomData.pixel_array
            rescale_slope = self.get_rescale_slope()
            rescale_intercept = self.get_rescale_intercept()

            resultArray = ( pixel_array * rescale_slope) + rescale_intercept
            return resultArray



        


    




    

    


import matplotlib.patches
from dicom_to_cnn.model.petctviewer.Roi import Roi


class RoiPolygon(Roi):
    """Derivated Class for manual Polygon ROI of PetCtViewer.org

    Returns:
        [RoiPolygon] -- Polygone ROI
    """

    def __init__(self, axis:int, first_slice:int, last_slice:int, roi_number:int, type_number:int, list_point:list, volume_dimension:tuple):
        """constructor

        Args:
            axis (int): [1 for axial, 2 for coronal, 3 for saggital]
            first_slice (int): [slice number where ROI begin]
            last_slice (int): [slice number where ROI end]
            roi_number (int): [roi number]
            type_number (int): [ 1 for axial polygone, 2 for coronal polygone, 3 for saggital polygone]
            list_point (list): [list of [x,y] coordonates of polygon in CSV]
            volume_dimension (tuple): [(shape x, shape y, shape z)]
        """
        super().__init__(axis, first_slice, last_slice, roi_number, type_number, list_point, volume_dimension)
        self.list_points = self.calculateMaskPoint()

    def calculateMaskPoint(self) -> list:
        """  calculate [x,y,z] coordonates/voxel which belong to the ROI in polygon patches

        Returns:
            [list]: [list of [x,y,z] coordonates of ROI]
        """
        roi_pixel_matplot = self.__create_closed_polygon()
        list_points = []
        for number_slice in range(self.first_slice - 1  , self.last_slice ) : 
            point = super().mask_roi_in_slice(roi_pixel_matplot)
            for i in range(len(point)):
                point[i].append(number_slice) 
            list_points.extend(point)
        return list_points

    def __create_closed_polygon(self) -> matplotlib.patches.Polygon:
        """generate an polygon patches from matplotlib

        Returns:
            [matplotlib.patches.Polygon]: [Polygon patche]
        """
        points_array = self.list_point_np
        return matplotlib.patches.Polygon(points_array, closed = True)


import SimpleITK as sitk 
import os


def convert_format_nifti_mask(nifti_path:str, destination:str):
    """Convert 8Bits UINT nifti mask to 32Bits UINT nifti mask

    Args:
        nifti_path ([str]): [path of the nifti file]
        destination ([str]): [path of destination's directory]
    """

    img = sitk.ReadImage(nifti_path)
    img = sitk.Cast(img, sitk.sitkUInt32)
    filename = nifti_path.split('/')[-1][0]
    sitk.WriteImage(img, os.path.join(destination,  filename))



import numpy as np
import matplotlib.patches 

class Roi():
    """A class to represent a ROI
    """

    def __init__(self, axis:int, first_slice:int, last_slice:int, roi_number:int, type_number:int, list_point:list, volume_dimension:tuple):
        """constructor

        Args:
            axis (int): [1 for axial, 2 for coronal, 3 for saggital]
            first_slice (int): [slice number where ROI begin]
            last_slice (int): [slice number where ROI end]
            roi_number (int): [roi number]
            type_number (int): [0 for nifti, 1 for axial polygone, 11 for axial ellipse, 2 for coronal polygone, 12 for coronal ellipse, 3 for saggital polygone, 13 for saggital ellipse]
            list_point (list): [list of [x,y] coordonates of polygone or ellipse / list of [x,y,z] coordonates of nifti]
            volume_dimension (tuple): [(shape x, shape y, shape z)]
        """
        self.axis = axis
        self.first_slice = first_slice
        self.last_slice = last_slice
        self.roi_number = roi_number
        self.type_number = type_number
        self.list_point = list_point
        self.list_point_np = np.asarray(self.list_point)
        self.x = volume_dimension[0]
        self.y = volume_dimension[1]
        self.z = volume_dimension[2]
        
    
    def __get_min_max_of_roi(self) -> tuple:
        """Compute extrema of ROI in which we will loop to find included voxel

        Arguments:
            point_list {np.ndarray} -- numpy point list

        Returns:
            [tuple] -- X/Y extremas
        """
        points_array = self.list_point_np
        all_x = points_array[:][:,0]
        all_y = points_array[:][:,1]
        if (self.type_number == 1 or self.type_number == 2 or self.type_number == 3) : #POLYGONE
            xmin = min(all_x)
            xmax = max(all_x)
            ymin = min(all_y)
            ymax = max(all_y)
            return xmin , xmax , ymin , ymax 
        else : #ELLIPSE
            height =  abs(all_x[0] - all_x[1])
            width = abs(all_y[0] - all_y[2])
            xmin = all_x[0] - height
            xmax = all_x[0] + height
            ymin = all_y[0] -  width
            ymax = all_y[0] + width 
            return xmin , xmax , ymin, ymax


    def mask_roi_in_slice(self, patch:matplotlib.patches) -> list:
        """get ROI x and y limits in which we will loop, to gather [x,y] pixel which are in the patch

        Args:
            patch (matplotlib.patches): [polygon or ellipse]]

        Returns:
            [list]: [list of [x,y] coordonates]
        """
        points = []
        xmin, xmax, ymin, ymax  = self.__get_min_max_of_roi()
        for x in range(xmin, xmax + 1): 
            for y in range(ymin, ymax + 1) : 
                if patch.contains_point([x,y], radius = -1e-9) : 
                    points.append([x,y])

        return points
    
    def get_empty_np_array(self) -> np.ndarray:
        """Return numpy array to fill given the current dimension and axis

        Returns:
            [np.ndarray] -- zero filled numpy array
        """
        return (np.zeros((self.x, self.y, self.z)))

    def coronal_to_axial(self, np_array_3D:np.ndarray) -> np.ndarray:
        """transform coronal 3d ndarray to 3d axial ndarray

        Args:
            np_array_3D (np.ndarray): [ROI ndarray]]

        Returns:
            [np.ndarray]: [return axial ndarray]
        """
        return np.transpose(np_array_3D, (2,1,0))

    def sagittal_to_axial(self, np_array_3D:np.ndarray) -> np.ndarray:
        """transform saggital 3d ndarray to 3d axial ndarray

        Args:
            np_array_3D (np.ndarray): [ROI ndarray]]

        Returns:
            [np.ndarray]: [return axial ndarray]
        """
        return np.transpose(np_array_3D, (0,2,1))

    def get_mask(self, list_points:list) -> np.ndarray : #list_points = [[x,y,z], [x,y,z], ...]
        """generate an empty ndarray and fill up with ROI coordonates

        Args:
            list_points (list): [ [[x,y,z], [x,y,z], [x,y,z], ...] ]

        Returns:
            [np.ndarray]: [return binary ndarray of the ROI]
        """
        np_array_3D = self.get_empty_np_array()
        for point in list_points:
            np_array_3D[point[1], point[0] , point[2]] = 1
        return np_array_3D.astype(np.uint8)



    

        
    









from dicom_to_cnn.model.reader.Series import Series
from dicom_to_cnn.model.reader.Instance_RTSS import Instance_RTSS
import cv2 as cv2
import numpy as np 

class MaskBuilder_RTSS : 
    """a class to build an ndarray mask from a RTSS File 
    """

    def __init__(self, rtss_path:str, serie_path:str):
        """constructor

        Args:
            rtss_path (str): [file path of RTSTRUCT dicom file]
            serie_path (str): [directory path of associated serie]
        """

        self.serie = Series(serie_path)
        self.serie.get_instances_ordered()
        self.serie_data = self.serie.get_series_details()
        self.instance_uid_serie = self.serie.get_all_SOPInstanceIUD()
        self.matrix_size = self.serie.get_size_matrix() #[x,y,z] from associated serie
        self.instances = self.serie.get_instances_ordered()
        self.image_position = self.instances[0].get_image_position()
        self.pixel_spacing = self.instances[0].get_pixel_spacing()
        self.pixel_spacing.append(self.serie.get_z_spacing())

        self.rtss = Instance_RTSS(rtss_path)
        self.number_of_roi = self.rtss.get_number_of_roi()


    def is_sop_instance_uid_same(self) -> bool:
        """check if every SOPInstanceUID from RTSTRUCT correspond to SOPINstanceUID from associated serie

        Returns:
            [bool]: [description]
        """
        uid_rtss = self.rtss.get_list_all_SOP_Instance_UID_RTSS()
        for uid in uid_rtss : 
            if uid not in self.instance_uid_serie : 
                return False

        return True

    def __spatial_to_pixels(self, number_roi:int) -> dict:
        """Transform spatial contour data to matrix coordonate contour

        Arguments :
            number_roi [(int)]: [a ROI number, start at 1]

        Returns:
            [dict] -- [{(roi)1 : { x : []
                               y : []
                               z : []} , 
                       {(roi)2 : { x : []
                               y : []
                               z : []} , ...]
        """
        
        list_contour_data = self.rtss.get_contour_data(number_roi) #x y z en mm 
        list_pixels = {}
        for i in range(3):
            self.image_position[i] = float(self.image_position[i])

        for contour_data in (list_contour_data):
            number_item = list_contour_data.index(contour_data) #0 1 2 3 ...
            contour_item = {}
            x = contour_data[::3]  #list
            x = [int(round((i - self.image_position[0]) / self.pixel_spacing[0] )) for i in x ] 
            contour_item['x'] = x
            y = contour_data[1::3] #list
            y = [int(round((i - self.image_position[1]) / self.pixel_spacing[1] ))  for i in y ]
            contour_item['y'] = y
            z = contour_data[2::3]
            z = [int(round((i - self.image_position[2]) / abs(self.pixel_spacing[2]) )) for i in z ][0]
            contour_item['z'] = z
            list_pixels[number_item + 1] = contour_item
        return list_pixels

    def get_list_points(self, number_roi:int) -> list:
        """transform a list of pixels of a ROI to a list nx2 (n points, coordonate (x,y)) for each contour and list of corresponding z slices

        Arguments:
            number_roi ([int]): [a ROI number, start at 1]
          

        Returns:
            [list] -- list of (x,y) points and list of corresponding z slices (in which there is a contour)
        """
        pixels = self.__spatial_to_pixels(number_roi) #dict 
        number_item = len(pixels)
        list_points = []
        slice = []
        for item in range(number_item):
            subliste = []
            list_x = (pixels[item + 1]['x'])
            list_y = (pixels[item + 1]['y'])
            for x,y in zip(list_x, list_y):
                subliste.append([x,y])

            list_points.append(subliste)

            slice.append(pixels[item + 1]['z'])
        return list_points, slice 
    

    def rtss_to_3D_mask(self, number_roi:int, matrix_size:list) -> np.ndarray:
        """method to generate 3D segmentation mask per ROI number

        Args:
            number_roi (int): [a ROI number, start at 1]
            matrix_size (list): [[shape x, shape y, shape z]]

        Raises:
            Exception: [raise Exception if one contour is not CLOSED_PLANAR ]

        Returns:
            [ndarray]: [return 3d ndarray of a ROI]
        """
        number_of_slices = matrix_size[2]
        np_array_3D = np.zeros(( matrix_size[0],  matrix_size[1], number_of_slices)).astype(np.uint8)
        if self.rtss.is_closed_planar(number_roi) == False : raise Exception ("Not CLOSED_PLANAR contour")
        liste_points, slice = self.get_list_points(number_roi)
        for item in range(len(slice)):
            np_array_3D[:,:,slice[item]] = cv2.drawContours(np.float32(np_array_3D[:,:,slice[item]]), [np.asarray(liste_points[item])], -1, number_roi , -1)
        return np_array_3D.astype(np.uint8)


    def rtss_to_4D_mask(self) -> np.ndarray:
        """method to generate 3d ndarray for each ROI, and stack them in a 4D ndarray

        Returns:
            [ndarray]: [return 4d ndarray of mask segmentation]
        """
        matrix_size = self.matrix_size
        np_array_3D = []
        for number_roi in range(1, self.number_of_roi +1):
            np_array_3D.append(self.rtss_to_3D_mask(number_roi, matrix_size))
        np_array_4D = np.stack((np_array_3D), axis = 3)
        return np_array_4D.astype(np.uint8)


import SimpleITK as sitk 
import abc 

class Abstract_Writer(metaclass = abc.ABCMeta) : 
    """abstract class to write segmentation
    """

    @abc.abstractmethod
    def save_file(self, filename:str, directory_path:str) -> None :
        pass 




import matplotlib.pyplot as plt
import numpy as np 
import scipy.ndimage
import os 
from fpdf import FPDF
import imageio 

class MIP_Generator : 
    """a class to generate MIP"""

    def __init__(self, numpy_array:np.ndarray):
        """constructor

        Args:
            numpy_array (np.ndarray): [3D np.ndarray of shape (z,y,x) or 4D np.ndarray of shape (z,y,x,c)]
        """
        self.numpy_array = numpy_array

    def project(self, angle:int) -> np.ndarray:
        """function to generate 2D MIP of a 3D (or 4D) ndarray of shape (z,y,x) (or shape (z,y,x,C)) 

        Args:
            angle (int): [angle of rotation of the MIP, 0 for coronal, 90 saggital ]

        Returns:
            [np.ndarray]: [return the MIP np.ndarray]
        """
        if len(self.numpy_array.shape) == 4 : 
            array = np.amax(self.numpy_array, axis = -1)
        else : 
            array = self.numpy_array
        axis = 1 
        vol_angle = scipy.ndimage.interpolation.rotate(array , angle=angle , reshape=False, axes = (1,2))
        MIP = np.amax(vol_angle,axis=axis)
        self.MIP = MIP
        return MIP


    def save_as_png(self, filename:str, directory:str, vmin:int=0, vmax:int=7) -> str:
        """method to save matplotlib.Figure of the generated MIP as png image

        Args:
            filename (str): [name of the image]
            directory (str): [directory's path where to save the new png image]
            vmin (int, optional): [minimum value of the MIP. If mask, vmin=None]. Defaults to 0.
            vmax (int, optional): [maximum value of the MIP, If mask, vmax=None]. Defaults to 7.

        Returns : 
            (str) : [return the abs path of the saved MIP]
        """
        filename = filename+'.png'
        f = plt.figure(figsize=(10,10))
        axes = plt.gca()
        axes.set_axis_off()
        if vmin is None or vmax is None : #mask
            plt.imshow(self.MIP, cmap = 'Reds', origin='lower')
        else : #pet 
            plt.imshow(self.MIP, cmap = 'gray', origin='lower', vmin = vmin, vmax = vmax)
        f.savefig(os.path.join(directory, filename), bbox_inches='tight')
        plt.close()
        return os.path.join(directory, filename)

    def create_mip_gif(self, filename:str, directory:str, vmin:int=0, vmax:int=7) -> None :
        """method to create mip GIF and save it as .gif

        Args:
            filename (str): [name of the gif]
            directory (str): [directory's path of the generated gif]
            vmin (int, optional): [mimimum value of the MIP]. Defaults to 0.
            vmax (int, optional): [maximum value of the MIP]. Defaults to 7.

        """
        duration = 0.1
        number_images = 60
        angle_filenames = []
        angles = np.linspace(0, 360, number_images)
        for angle in angles:
            MIP = self.project(angle)
            mip_filename=str(angle)+'.png'
            path = self.save_as_png(mip_filename, directory, vmin, vmax)
            angle_filenames.append(path)
        self.files_to_gif(angle_filenames, duration, filename, directory)
        for image in angle_filenames : 
            os.remove(image)

    @classmethod
    def files_to_gif(cls, filenames:list, duration:float, name:str, directory:str) -> None :
        """From a list of images, create gif

        Args:
            filenames ([list]): [list of all images' path]
            duration ([float]): [time of each image]
            name ([str]): [gif name]
            directory ([str]): [gif directory]
        """
        images = []
        for filename in filenames:
            images.append(imageio.imread(filename))
        output_file = directory+'/' + name +'.gif'
        imageio.mimwrite(output_file, images, duration=duration)

    @classmethod
    def create_pdf_mip(cls, angle_filenames:list, filename:str, directory:str)-> None : 
        """function generate pdf file of PET MIP and MASK MIP 
        
            Arguments : 
            angle_filenames ([list]) : [list of mip path and study_uid : [path_mip_pet, path_mip_mask, title], [path_mip_pet, path_mip_mask, title],... ]
            filename ([str]) : [name of the pdf file]
            directory ([str]) : [directory's path where to save the pdf file]
        """
        pdf = FPDF()
        for mip in angle_filenames : 
            pdf.add_page()
            pdf.image(mip[0], x = 0, y = 10, w = 100, h = 190)
            pdf.image(mip[1], x = 100, y = 10, w = 100, h = 190)
            pdf.set_font("Arial", size=12)
            pdf.cell(200, 0, txt= str(mip[2]), ln=2, align="C")
        pdf.output(os.path.join(directory, filename))


    

    



import numpy as np 
import SimpleITK as sitk 
import os

#target_size = (128, 128, 256)
#target_spacing = (4.0, 4.0, 4.0)
#target_direction = (1,0,0,0,1,0,0,0,1)


class Fusion : 

    def __init__(self):
        pass 

    def set_origin_image(self, origin_img:sitk.Image) -> None : 
        """method to set the origin sitk.Image on which we want to resample 

        Args:
            origin_img (sitk.Image): []
        """
        self.origin_img = origin_img 
        self.origin_size = origin_img.GetSize()
        self.origin_spacing = origin_img.GetSpacing()
        self.origin_direction = origin_img.GetDirection()
        self.origin_origin = origin_img.GetOrigin()

    def set_target_volume(self, target_size: tuple, target_spacing: tuple, target_direction:tuple) -> None : 
        """method to set the target size, spacing and direction 

        Args:
            target_size (tuple): [(x,y,z)]
            target_spacing (tuple): [(x,y,z)]
            target_direction (tuple): [(x,x,x,y,y,y,z,z,z)]
        """
        self.target_size = target_size 
        self.target_spacing = target_spacing 
        self.target_direction = target_direction 

    def __compute_new_origin_head2hips(self) -> None :
        """method to compute new_origin from head

        Returns:
            [tuple]: [return new origin]
        """
        origin_origin = np.asarray(self.origin_origin)
        origin_size = np.asarray(self.origin_size)
        origin_spacing = np.asarray(self.origin_spacing)
        new_origin = (origin_origin[0] + 0.5 * origin_size[0] * origin_spacing[0] - 0.5 * self.target_size[0] * self.target_spacing[0],
                      origin_origin[1] + 0.5 * origin_size[1] * origin_spacing[1] - 0.5 * self.target_size[1] * self.target_spacing[1],
                      origin_origin[2] + 1.0 * origin_size[2] * origin_spacing[2] - 1.0 * self.target_size[2] * self.target_spacing[2])
        return new_origin 
        
    def resample(self, image_to_resample:sitk.Image, defaultValuePixel:float) -> sitk.Image:
        """method to resample a sitk.Image with defined size, target, spacing and origin

        Args:
            image_to_resample (sitk.Image): [shape (x,y,z)]
            defaultValuePixel (float): [Set the pixel value when a transformed pixel is outside of the image, pet and mask =0.0, ct = -1000.0]

        Returns:
            sitk.Image: [return the resampled sitk.Image]
        """
        target_origin = self.__compute_new_origin_head2hips()
        transformation = sitk.ResampleImageFilter()
        transformation.SetOutputDirection(self.target_direction)
        transformation.SetOutputOrigin(target_origin)
        transformation.SetOutputSpacing(self.target_spacing)
        transformation.SetSize(self.target_size)
        transformation.SetDefaultPixelValue(defaultValuePixel)
        transformation.SetInterpolator(sitk.sitkLinear)
        resampled_img = transformation.Execute(image_to_resample)
        return resampled_img 




    



import os 
import numpy as np
import matplotlib.pyplot as plt
from dicom_to_cnn.model.post_processing.mip.MIP_Generator import MIP_Generator

def transform_to_nan(MIP:np.ndarray) -> np.ndarray:
    """function to replace pixel with value 0 with np.NaN

    Args:
        MIP (np.ndarray): [2D np.ndarray of MIP]

    Returns:
        (np.ndarray): [2D np.ndarray NaN of MIP]
    """
    nan_MIP = np.empty(MIP.shape)
    nan_MIP[:] = np.NaN
    y,x = np.where(MIP>0)
    nan_MIP[y,x] = MIP[y,x]
    return nan_MIP 


def save_projection_two_modality(pet_array:np.ndarray, mask_array:np.ndarray, angle:int, filename:str, directory:str, vmin:int=0, vmax:int=7) -> str :
    """function to generate a MIP of PET/MASK and save it as png image

    Args:
        pet_array (np.ndarray): [3D np.ndarray of shape (z,y,x)]
        mask_array (np.ndarray): [3D np.ndarray of shape (z,y,x)]
        angle (int): [angle of the MIP rotation]
        filename (str): [filename of the png image]
        directory (str): [directory's path where to save the png image]
        vmin (int, optional): [minimum value of the PET MIP]. Defaults to 0.
        vmax (int, optional): [maximum value of the PET MIP]. Defaults to 7.

    Returns:
        [str]: [return the complete abs path to the generated PNG]
    """
    mip_pet = MIP_Generator(pet_array).projection(angle)
    mip_mask = MIP_Generator(mask_array).projection(angle)
    mip_mask = transform_to_nan(mip_mask)
    f = plt.figure(figsize=(10,10))
    axes = plt.gca()
    axes.set_axis_off() 
    plt.imshow(mip_pet, vmin=vmin, vmax=vmax, cmap='Greys', origin='lower')
    plt.imshow(mip_mask, cmap='hsv', alpha = 0.5, origin='lower')
    f.savefig(os.path.join(directory, filename+'.png'), bbox_inches='tight')
    plt.close()
    return os.path.join(directory, filename+'.png')

def create_gif_two_modality(pet_array:np.ndarray, mask_array:np.ndarray, filename:str, directory:str, vmin:int=0, vmax:int=7):
    """function to generate a gif MIP of PET/MASK and save it as .gif

    Args:
        pet_array (np.ndarray): [3D np.ndarray of shape (z,y,x)]
        mask_array (np.ndarray): [3D np.ndarray of shape (z,y,x)]
        filename (str): [filename of the png image]
        directory (str): [directory's path where to save the png image]
        vmin (int, optional): [minimum value of the PET MIP]. Defaults to 0.
        vmax (int, optional): [maximum value of the PET MIP]. Defaults to 7.

    """
    duration = 0.1
    number_images = 60
    angle_filenames = []
    angles = np.linspace(0, 360, number_images)
    for angle in angles:
        mip_filename = str(angle)
        angle_filenames.append(save_projection_two_modality(pet_array, mask_array, angle, mip_filename, directory, vmin, vmax))
    MIP_Generator.create_gif(angle_filenames, duration, filename, directory)
    for image in angle_filenames : 
        os.remove(image)







import pydicom 
import numpy as np
from random import randrange
from skimage.measure import label, find_contours
import SimpleITK as sitk 

class ROIContourSequence : 
    """a class to represente ROIContourSequence in RTSTRUCT file
    """

    def __init__(self, mask_array:np.ndarray, mask_img:sitk.Image, number_of_roi:int):
        """constructor

        Args:
            mask_array (np.ndarray]): [mask ndarray]
            mask_img ([sitk.Image]): [mask sitk image]
            number_of_roi ([int]): [total number of roi ]
        """
        self.mask_array = mask_array 
        self.mask_img = mask_img
        self.number_of_roi = number_of_roi


    def __define_contour_img(self, number_roi:int) -> sitk.Image :
        """method to extract mask img per ROI number

        Args:
            number_roi (int): [A roi number, start at 1]

        Returns:
            [sitk.Image]: [mask img of ROI number 'number_roi']
        """
        roi = np.zeros((self.mask_array.shape), dtype=np.uint8)
        z,x,y = np.where(self.mask_array == number_roi)
        roi[z,x,y] = 1
        roi_img = sitk.GetImageFromArray(roi)
        return roi_img 


    def get_spatial_coordonate(self, number_roi:int, list_all_SOPInstanceUID:list) -> tuple:
        """Per ROI number, gather spatial coordonates per slices (when there is a contour)
        Args:
            number_roi (int): [a ROI number, start at 1]
            list_all_SOPInstanceUID (list): [list of every SOPInstanceUID from associated dicom serie]

        Returns:
            [tuple]: [list of spatial coordonates [ [contour 1 : [x, y, z, x, y, z...] ], [contour 2 : [x, y, z, x, y, z...]], ... ] and list of SOPInstanceUID [SOPInstanceUID_contour1, SOPINStanceUID_contour2, ...]

        """
        img_roi = self.__define_contour_img(number_roi)
        depth = img_roi.GetDepth()
        results = []
        list_SOPInstance = []
        for z in range(depth):
            img_slice = img_roi[:,:,z]
            array_slice = sitk.GetArrayFromImage(img_slice)
            contour = find_contours(array_slice, level = 0.0)
            if contour != []: 
                for i in range(len(contour)):
                    liste = []
                    l = contour[i].tolist()
                    for item in l:
                        spatial_coord = self.mask_img.TransformIndexToPhysicalPoint([int(item[1]), int(item[0]), int(z)])
                        liste.append(spatial_coord[0])
                        liste.append(spatial_coord[1])
                        liste.append(spatial_coord[2])

                    results.append(liste)
                    list_SOPInstance.append(list_all_SOPInstanceUID[z])

        return results, list_SOPInstance 

    

    def __create_ContourImageSequence(self, ReferencedSOPClassUID:str, ReferencedSOPInstanceUID:str) -> pydicom.Sequence:
        """method to generate ContourImageSequence from ROIContourSequence 

        Args:
            ReferencedSOPClassUID (str): [Referenced SOP Class UID value from associated serie]
            ReferencedSOPInstanceUID (str): [Reference SOPInstance UID value from associated serie]

        Returns:
            [pydicom.Sequence]: [return ContourImageSequence]
        """
        ContourImageSequence = pydicom.sequence.Sequence()
        dataset = pydicom.dataset.Dataset()
        dataset.ReferencedSOPClassUID = ReferencedSOPClassUID
        dataset.ReferencedSOPInstanceUID = ReferencedSOPInstanceUID
        ContourImageSequence.append(dataset)
        return ContourImageSequence 

    def __create_ContourSequence(self, ReferencedSOPClassUID:str, list_ReferencedSOPInstanceUID:list, list_ContourData:list) -> pydicom.Sequence:
        """method to generate ContourSequence from ROIContourSequence

        Args:
            ReferencedSOPClassUID (str): [Referenced SOP Class UID value from associated serie]
            list_ReferencedSOPInstanceUID (list): [list of every SOPInstanceUID (in which we find contour), same size as list_ContourData]
            list_ContourData (list): [list of every ContourData [[x,y,z,x,y,z], [x,y,z,...], ...], same size as list_ReferencedSOPInstanceUID]


        Returns:
            [pydicom.Sequence]: [return ContourSequence]
        """
        ContourSequence = pydicom.sequence.Sequence()

        for ContourData,SOPInstanceUID in zip(list_ContourData,list_ReferencedSOPInstanceUID):
            dataset = pydicom.dataset.Dataset()
            dataset.ContourData = ContourData             
            dataset.ContourImageSequence = self.__create_ContourImageSequence(ReferencedSOPClassUID, SOPInstanceUID)             
            dataset.NumberOfContourPoints = len(ContourData)/3 
            dataset.ContourGeometricType = 'CLOSED_PLANAR'

            ContourSequence.append(dataset)
        return ContourSequence 


    @classmethod
    def get_random_colour(cls) -> list:
        """a class method to generate random color for ROI

        Returns:
            [list]: [return color [r,g,b]]
        """
        max = 256
        return [randrange(max), randrange(max), randrange(max)]


    def create_ROIContourSequence(self, ReferencedSOPClassUID:str, list_all_SOPInstanceUID:list) -> pydicom.Sequence:
        """method to generate ROIContourSequence from RTSTRUCT file 

        Args:
            ReferencedSOPClassUID (str): [ReferencedSOPClass UID value from associated serie]
            list_all_SOPInstanceUID (list): [list of every SOPInstanceUID of each instance from associated dicom serie]

        Returns:
            [pydicom.Sequence]: [return ROIContourSequence]
        """

        ROIContourSequence = pydicom.sequence.Sequence()
        for number_roi in range(1, self.number_of_roi +1) : 
            dataset = pydicom.dataset.Dataset()
            dataset.ROIDisplayColor = self.get_random_colour()
            dataset.ReferencedROINumber = number_roi
            list_contour_data, list_SOP_instance_uid = self.get_spatial_coordonate(number_roi, list_all_SOPInstanceUID )
            dataset.ContourSequence = self.__create_ContourSequence(ReferencedSOPClassUID, list_SOP_instance_uid, list_contour_data)
            ROIContourSequence.append(dataset)
        return ROIContourSequence 





import pydicom
import datetime
import random 
import os 
import numpy as np 
import tempfile 
import SimpleITK as sitk 
from skimage.measure import label 
from dicom_to_cnn.model.segmentation.Abstract_Writer import Abstract_Writer
from dicom_to_cnn.model.reader.Series import Series
from dicom_to_cnn.model.segmentation.rtstruct.StructureSetROISequence import StructureSetROISequence
from dicom_to_cnn.model.segmentation.rtstruct.RTROIObservationsSequence import RTROIObservationsSequence
from dicom_to_cnn.model.segmentation.rtstruct.ROIContourSequence import ROIContourSequence
from dicom_to_cnn.model.segmentation.rtstruct.ReferencedFrameOfReferenceSequence import *
from dicom_to_cnn.tools.export_segmentation.generate_dict import *


class RTSS_Writer(Abstract_Writer):
    """A class to write a DICOM RTSTRUCT file
    """
    
    def __init__(self, mask_img:sitk.Image, serie_path:str):
        """constructor
        Args:
            mask ([sitk.Image]): [3D sitk.Image of segmentation (x,y,z), labelled or not]
            serie_path ([str]): [Serie path related to RTSTRUCT file ]
        """
        #SERIE
        self.mask_img = mask_img#GetSize() = [x,y,z]
        serie = Series(serie_path)
        self.instances = serie.get_instances_ordered()

        #Get list of every sop instance uid from associated serie
        self.list_all_SOPInstanceUID = serie.get_all_SOPInstanceIUD()

        #MASKs
        self.mask_array = sitk.GetArrayFromImage(self.mask_img) #shape = [z,x,y]
        self.mask_array = self.__clean_mask()
        self.image_position = self.mask_img.GetOrigin()
        self.pixel_spacing = self.mask_img.GetSpacing()
        self.image_direction = self.mask_img.GetDirection()

        #get number of ROI 
        self.number_of_roi = int(np.max(self.mask_array))


    def setDictName(self, name:str) -> str : 
        """set the dict name 

        Args:
            name (str): [dict name]

        """
        self.dict_name = name

    def setBodyPartExaminated(self, name:str) -> str : 
        """set the body part examinated

        Args:
            name (str): [name of the body part]

        """
        if len(name) > 16 : self.body_part =  name[0:16]
        else : self.body_part = name 

    def setSeriesDescription(self, description:str) -> str : 
        """set the study description of the exam

        Args:
            description (str): [study description]


        """
        if len(description) > 16 : self.serie_description =  description[0:16]
        else : self.serie_description = description 

    def setRTROIInterpretedType(self, dictionnary:dict) -> dict : 
        """set a dict for the RTROIInterpretedType value for each ROI

        Args:
            dictionnary (dict): [{1 : value,
                                  2 : value, 
                                  etc }]

        """ 
        self.roi_type_dict = dictionnary

    def setAutoRTROIInterpretedType(self)-> dict : 
        """set an auto dict with RTROIINterpretedType value 

            generated dict  {1 : '', 
                            2 : '', 
                            etc }]
        """
        number_of_roi = int(np.max(self.mask_array))
        dictionnary = {}
        for i in range(1, number_of_roi+1):
            dictionnary[i] = str('')
        self.roi_type_dict = dictionnary

    def setRoiName(self, dictionnary:dict) -> dict : 
        """set a dict for ROI Name, for each ROI

        Args:
            dictionnary (dict): [{1 : 'name roi 1', 
                                  2:'name roi 2', 
                                 etc }]

        """
        self.roi_name_dict = dictionnary

    def setAutoRoiName(self) -> dict : 
        """set an auto dict with ROI Name value 

            generated dict  {1 : 'ROI 1', 
                            2 : 'ROI 2', 
                                etc }]
        """
        mask_array = sitk.GetArrayFromImage(self.mask_img)
        number_of_roi = int(np.max(mask_array))
        dictionnary = {}
        for i in range(1, number_of_roi+1):
            dictionnary[i] = str("ROI {}".format(i+1))
        self.roi_name_dict =  dictionnary

    def __generate_dict_json(self) -> str:
        results = generate_dict(self.number_of_roi, self.dict_name, self.serie_description, self.body_part, self.roi_name_dict, interpreted_type=self.roi_type_dict)
        self.results = results

    def __clean_mask(self) -> np.ndarray:
        """a function to clean mask : remove pixel (=1) when there is less than 3 pixels (=1) in slice

        Args:
            mask (np.ndarray): [mask of shape [z,y,x]]

        Returns:
            [ndarray]: [return the cleaned mask in shape [z,y,x]]
        """
        number_of_roi = int(np.max(self.mask_array))
        binary_mask = np.where(self.mask_array>0, 1, 0)
        
        empty_mask = np.zeros((self.mask_array.shape))
        for s in range(self.mask_array.shape[0]) : 
            slice = binary_mask[s, :, :]
            if int(np.max(binary_mask[s,:,:])) == 0 : 
                empty_mask[s, :, :] = slice
            else : 
                lw, num = label(slice, connectivity=2, return_num=True) #lw = 2D slice 
                item = np.arange(1, num+1).tolist()
                area = []
                for it in item : 
                    area.append(len(np.where(lw== it)[0]))
                for ar in area : 
                    feature = area.index(ar) + 1 
                    if int(ar) < 3 : 
                        x,y = np.where(lw == feature)
                        lw[x,y] = 0 
                empty_mask[s, :, :] = lw 
        matrix = np.where(empty_mask>0, 1, 0)
        #labelled again 
        label_ = 1
        for i in range(1, number_of_roi+1):
            z,y,x = np.where((matrix > 0) & (self.mask_array == i))
            if len(z) == 0 : 
                pass 
            else : 
                matrix[np.where((matrix > 0) & (self.mask_array == i))] = label_ 
                label_ += 1 
        return matrix

    def generates_file_meta(self) -> tuple:
        """
            Generates required values for file meta information
            List of tags :
                  - FileMetaInformationGroupLength':
                  - FileMetaInformationVersion'    :
                  - MediaStorageSOPClassUID'       :
                  - MediaStorageSOPInstanceUID'    :
                  - TransferSyntaxUID'             :
                  - ImplementationClassUID'        :

        Returns: 
            (tuple): [return file_meta and MediaStorageSOPInstanceUID of the new RTSTRUCT]
        """


        file_meta = pydicom.dataset.FileMetaDataset()
        file_meta.FileMetaInformationGroupLength = 166
        file_meta.FileMetaInformationVersion = b'\x00\x01'
        file_meta.MediaStorageSOPClassUID = '1.2.840.10008.5.1.4.1.1.481.3' # RT Structure Set Storage
        file_meta.MediaStorageSOPInstanceUID = pydicom.uid.generate_uid()
        file_meta.TransferSyntaxUID = pydicom.uid.ExplicitVRLittleEndian #pydicom.uid.ImplicitVRLittleEndian 
        file_meta.ImplementationClassUID =  pydicom.uid.PYDICOM_IMPLEMENTATION_UID #'1.2.246.352.70.2.1.7'
        
        return file_meta, file_meta.MediaStorageSOPInstanceUID


    def set_tags(self) -> None:
        """ generate required values from associated dicom serie
            
            List of tags :
                  - AccessionNumber'       :
                  - PatientBirthDate'      :
                  - PatientBirthTime'      :
                  - PatientID'             :
                  - PatientName'           :
                  - PatientSex'            :
                  - PhysiciansOfRecord'    :
                  - ReferringPhysicianName':
                  - SpecificCharacterSet'  :
                  - StudyDate'             :
                  - StudyDescription'      :
                  - StudyID'               :
                  - StudyInstanceUID'      :
                  - StudyTime'             : 
        """     
        #from the serie 
        self.dataset.AccessionNumber = self.instances[0].get_accession_number()
        self.dataset.PatientBirthDate = self.instances[0].get_patient_birth_date()
        self.dataset.PatientID = self.instances[0].get_patient_id()
        self.dataset.PatientName = self.instances[0].get_patient_name()
        self.dataset.PatientSex = self.instances[0].get_patient_sex()
        self.dataset.PhysiciansOfRecord = self.instances[0].get_physicians_of_record()
        self.dataset.ReferringPhysicianName = self.instances[0].get_referring_physician_name()
        self.dataset.SpecificCharacterSet = self.instances[0].get_specific_character_set()
        self.dataset.StudyDate = self.instances[0].get_study_date()
        self.dataset.StudyDescription = self.instances[0].get_study_description()
        self.dataset.StudyID = self.instances[0].get_study_id()
        self.dataset.StudyInstanceUID = self.instances[0].get_study_instance_uid()
        self.dataset.StudyTime = self.instances[0].get_study_time()

        #specific new tags for the serie 
        self.dataset.ApprovalStatus = 'UNAPPROVED'
        self.dataset.Manufacturer   = ''
        dt = datetime.datetime.now()
        self.dataset.InstanceCreationDate = dt.strftime('%Y%m%d')
        self.dataset.InstanceCreationTime = dt.strftime('%H%M%S.%f')
        self.dataset.InstanceNumber = '1'
        self.dataset.Modality = 'RTSTRUCT'
        self.dataset.ReviewDate = '' #because UNAPPROVED
        self.dataset.ReviewTime = '' #because UNAPPROVED
        self.dataset.ReviewerName = '' #because UNAPPROVED
        self.dataset.SeriesDescription = self.results["SeriesDescription"]
        self.dataset.SeriesInstanceUID = pydicom.uid.generate_uid()
        self.dataset.SeriesNumber = random.randint(0,1e3)
        self.dataset.SOPClassUID = '1.2.840.10008.5.1.4.1.1.481.3' 
        self.dataset.SOPInstanceUID = self.file_meta.MediaStorageSOPInstanceUID
        self.dataset.StructureSetDate = dt.strftime('%Y%m%d')
        self.dataset.StructureSetDescription = self.results["SeriesDescription"]
        self.dataset.StructureSetLabel = self.results["SeriesDescription"]
        self.dataset.StructureSetTime = dt.strftime('%H%M%S.%f')

    #StructureSetROISequence
    def set_StructureSetROISequence(self) -> None :
        """method to set StructureSetROISequence from RTSTRUCT file 
        """
        referenced_frame_of_reference_uid = self.instances[0].get_frame_of_reference_uid()
        self.dataset.StructureSetROISequence = StructureSetROISequence(self.mask_array, self.results, self.number_of_roi).create_StructureSetROISequence(self.pixel_spacing, referenced_frame_of_reference_uid)
        
    #RTROIObservationSequence
    def set_RTROIObservationSequence(self) -> None :
        """method to set RTROIObservationSequence from RTSTRUCT file
        """
        self.dataset.RTROIObservationsSequence = RTROIObservationsSequence(self.results, self.number_of_roi).create_RTROIObservationsSequence()
        
    #ROIContourSequence 
    def set_ROIContourSequence(self)-> None:
        """method to set ROIContourSequence from RTSTRUCT file 
        """
        referenced_sop_class_uid = self.instances[0].get_sop_class_uid()
        self.dataset.ROIContourSequence = ROIContourSequence(self.mask_array, self.mask_img, self.number_of_roi).create_ROIContourSequence(referenced_sop_class_uid, self.list_all_SOPInstanceUID)

    #ReferencedFrameOfReferenceSequence 
    def set_ReferencedFrameOfReferenceSequence(self) -> None:
        """method to set ReferencedFrameOfReferenceSequence from RTSTRUCT file
        """
        frame_of_reference_uid = self.instances[0].get_frame_of_reference_uid()
        series_instance_uid = self.instances[0].get_series_instance_uid()
        study_instance_uid = self.instances[0].get_study_instance_uid()
        sop_class_uid = self.instances[0].get_sop_class_uid()
        self.dataset.ReferencedFrameOfReferenceSequence = create_ReferencedFrameOfReferenceSequence(frame_of_reference_uid, sop_class_uid, self.list_all_SOPInstanceUID, series_instance_uid, study_instance_uid)


    def rtstruct_writer(self):
        #generate dictionnary with parameter inside from 'generate_dict.py'
        self.__generate_dict_json()

        #RTSTRUCT FILE 
        #creation file_meta 
        self.file_meta, self.file_meta.MediaStorageSOPInstanceUID = self.generates_file_meta()
        suffix = '.dcm'
        filename_little_endian = tempfile.NamedTemporaryFile(suffix=suffix).name
        self.dataset = pydicom.dataset.FileDataset(filename_little_endian, {}, file_meta=self.file_meta, preamble=b"\0" * 128)

        #Add the data element in FileDataset
        self.set_tags()
        self.set_StructureSetROISequence()
        self.set_RTROIObservationSequence()
        self.set_ROIContourSequence()
        self.set_ReferencedFrameOfReferenceSequence()
        
        # Set the transfer syntax
        self.dataset.is_little_endian = True
        self.dataset.is_implicit_VR = False 

    def save_file(self, filename:str, directory_path:str) -> None :
        """method to save the RTSTRUCT file

        Args:
            filename (str): [name of the RTSTRUCT file]
            directory_path (str): [directory's path where to save the RTSTRUCT file]

        """
        self.rtstruct_writer()
        self.dataset.save_as(os.path.join(directory_path, filename), write_like_original=False)


        
        

    



import numpy as np
import SimpleITK as sitk 
import matplotlib.pyplot as plt
from skimage.feature import peak_local_max
from skimage.segmentation import watershed
import scipy.ndimage
from skimage import morphology


class Watershed:
    """A class to generate subsegmentation mask with a watershed model
    """

    def __init__(self, binary_img:sitk.Image, pet_img:sitk.Image):
        """constructor

        Args:
            binary_img (sitk.Image): [binary segmentation sitk.Image (x,y,z) ]
            pet_img (sitk.Image): [pet sitk.Image associated (x,y,z)]
        """
        self.binary_img = binary_img
        self.pet_img = pet_img
        self.binary_array = sitk.GetArrayFromImage(self.binary_img) #(z,y,x)
        self.pet_array = sitk.GetArrayFromImage(self.pet_img) # (z,y,x)
        self.generated_connected_component_img()

    def generated_connected_component_img(self) -> None :
        """method to generate a first labelled img and its np.ndarray by connected component
        """
        self.labelled_img = sitk.ConnectedComponent(self.binary_img)
        self.labelled_array = sitk.GetArrayFromImage(self.labelled_img) #(z,y,x)
        self.number_of_cc = int(np.max(self.labelled_array))

    def get_suv_values_matrix(self, label:int) -> np.ndarray:
        """method to get a new SUV ndarray for a chosen label

        Args:
            label ([int]): [a chosen label]

        Returns:
            [np.ndarray]: [description]
        """
        new_matrix_pet = np.zeros(self.labelled_array.shape) #(z,y,x)
        new_matrix_pet[np.where(self.labelled_array == label)] = self.pet_array[np.where(self.labelled_array== label)]
        return new_matrix_pet

    def get_mask_roi_matrix(self, label:int) -> np.ndarray:
        """method to get a new ROI ndarray for a chosen label

        Args:
            label (int): [a choosen label]

        Returns:
            [np.ndarray]: [description]
        """
        new_matrix_roi = np.zeros(self.labelled_array.shape)
        new_matrix_roi[np.where(self.labelled_array == label)] = 1
        return new_matrix_roi.astype(np.uint8)

    def get_distance_map(self, array:np.ndarray) -> np.ndarray:
        """generate distance map of a given array

        Args:
            array (np.ndarray): [np.ndarray (z,y,x)]

        Returns:
            [np.ndarray]: [return distance map of the given array]
        """
        return scipy.ndimage.distance_transform_edt(array)

    def get_local_peak(self, distance_map:np.ndarray) -> np.ndarray:
        """calculate local peak from a given distance map (np.ndarray)

        Args:
            distance_map (np.ndarray): [a 3D ndarray/distance map (z,y,x)]

        Returns:
            [np.ndarray]: [return a np.ndarray with local peak points (= 1)]
        """
        spacing = []
        pet_spacing = self.pet_img.GetSpacing()
        spacing.append(float(pet_spacing[0]) * 10**(-1))
        spacing.append(float(pet_spacing[1]) * 10**(-1))
        spacing.append(float(pet_spacing[2]) * 10**(-1))     
        min_dist = int(2/np.mean(spacing))
        return peak_local_max(distance_map, indices = False, min_distance=min_dist)


    def define_marker_array(self, localMax:np.ndarray) -> tuple:
        """method to attribute a label on every local peak points

        Args:
            localMax (np.ndarray): [a np.ndarray with local peak points (z,y,x)]

        Returns:
            [tuple]: [return a labelled local peak points np.ndarray and the number of local peak points]
        """
        marker_array, num_features = scipy.ndimage.label(localMax) 
        return marker_array.astype(np.uint8), num_features
 

    def applied_watershed_model(self) -> sitk.Image : 
        """applied watershed model process

        Returns:
            [sitk.Image]: [return the watershed labelled sitk.Image]
        """
        new_coordonates = []
        pet_spacing = self.pet_img.GetSpacing()
        volume_voxel = pet_spacing[0] * pet_spacing[1] * pet_spacing[2] * 10**(-3) #in ml 
        labels = np.arange(1, self.number_of_cc+1, 1)
        for label in labels: 
            pixel_roi = int(len(np.where(self.labelled_array == label)[0]))
            volume_roi = pixel_roi * volume_voxel 
            if volume_roi < int(30) : 
                new_coordonates.append(np.where(self.labelled_array == label))
            else : 
                suv_values= self.get_suv_values_matrix(label)
                localMax = self.get_local_peak(suv_values) 
                marker_array, num_features = self.define_marker_array(localMax)
                if num_features != 0 : 
                    new_label_mask = watershed(image=-suv_values, markers=marker_array, mask=suv_values)
                    new_label_mask = new_label_mask.astype(np.uint8)
                    for new_label in range(1, num_features + 1):
                        if len(np.where(new_label_mask == new_label)[0]) != 0 : 
                            new_coordonates.append(np.where(new_label_mask == new_label))
            
        number_total_of_label = len(new_coordonates)
        watershed_array = np.zeros(self.binary_array.shape)
        for coordonate, label in zip(new_coordonates, np.arange(1, number_total_of_label+1, 1)):
            watershed_array[coordonate] = label 
        pet_spacing = self.pet_img.GetSpacing()
        pet_direction = self.pet_img.GetDirection()
        pet_origin = self.pet_img.GetOrigin()

        watershed_img = sitk.GetImageFromArray(watershed_array)
        watershed_img.SetSpacing(pet_spacing)
        watershed_img.SetOrigin(pet_origin)
        watershed_img.SetDirection(pet_direction)
        
        return watershed_img
        


    @classmethod
    def remove_small_roi(cls, binary_img:sitk.Image, pet_img:sitk.Image) -> sitk.Image:
        """function to remove ROI under 30 ml on a binary sitk.Image

        Args:
            binary_img (sitk.Image): [sitk.Image of size (z,y,x)]
            pet_img (sitk.Image): [sitk.Image of the PET, size (z,y,x)]

        Raises:
            Exception: [raise Exception if not a 3D binary mask]

        Returns:
            [sitk.Image]: [Return cleaned image]
        """

        binary_array = sitk.GetArrayFromImage(binary_img)
        if len(binary_array.shape) != 3 or int(np.max(binary_array)) != 1 : 
            raise Exception("Not a 3D binary mask, need to transform into 3D binary mask")
        else : 
            pet_spacing = pet_img.GetSpacing()
            pet_origin = pet_img.GetOrigin()
            pet_direction = pet_img.GetDirection()
            labelled_img = sitk.ConnectedComponent(binary_img)
            stats = sitk.LabelIntensityStatisticsImageFilter()
            stats.Execute(labelled_img, pet_img)
            labelled_array = sitk.GetArrayFromImage(labelled_img).transpose()
            number_of_label = stats.GetNumberOfLabels()
            volume_voxel = pet_spacing[0] * pet_spacing[1] * pet_spacing[2] * 10**(-3) #in ml 
            for i in range(1, number_of_label + 1) :
                volume_roi = stats.GetNumberOfPixels(i) * volume_voxel
                if volume_roi < float(30) : 
                    x,y,z = np.where(labelled_array == i)
                    for j in range(len(x)):
                        labelled_array[x[j], y[j], z[j]] = 0
            new_binary_array = np.zeros((labelled_array.shape))
            new_binary_array[np.where(labelled_array != 0)] = 1
            new_binary_img = sitk.GetImageFromArray(new_binary_array.transpose().astype(np.uint8))
            new_binary_img.SetOrigin(pet_origin)
            new_binary_img.SetSpacing(pet_spacing)
            new_binary_img.SetDirection(pet_direction)
            return new_binary_img



import matplotlib.patches
import numpy as np
import math
from dicom_to_cnn.model.petctviewer.Roi import Roi


class RoiElipse(Roi):
    """Derivated Class for manual Elipse ROI of PetCtViewer.org

    Returns:
        [RoiElipse] -- Roi Elipse Object
    """

    def __init__(self, axis:int, first_slice:int, last_slice:int, roi_number:int, type_number:int, list_point:list, volume_dimension:tuple):
        """constructor

        Args:
            axis (int): [1 for axial, 2 for coronal, 3 for saggital]
            first_slice (int): [slice number where ROI begin]
            last_slice (int): [slice number where ROI end]
            roi_number (int): [roi number]
            type_number (int): [ 11 for axial ellipse, 12 for coronal ellipse, 13 for saggital ellipse]
            list_point (list): [list of [x,y] coordonates of ellipse in CSV]
            volume_dimension (tuple): [(shape x, shape y, shape z)]
        """
        super().__init__(axis, first_slice, last_slice, roi_number, type_number, list_point, volume_dimension)
        self.list_points = self.calculateMaskPoint()


    def calculateMaskPoint(self):
        """ calculate [x,y,z] coordonates/voxel which belong to the ROI in ellipse patches

        Returns:
            [list]: [list of [x,y,z] coordonates of ROI ]
        """
        points_array = self.list_point_np 
        x0 = points_array[0][0]
        y0 = points_array[0][1]
        delta_x = points_array[1][0] - x0
        delta_y = abs(points_array[2][1] - y0)
        middle = ((self.first_slice + self.last_slice) / 2) - 1
        rad1 = ((self.last_slice - self.first_slice) / 2) 
        list_points = []
        if(rad1 == 0) :
            return list_points
        for number_slice in range(self.first_slice  , self.last_slice - 1 ) : 
            diff0 = abs(middle - number_slice)
            factor = pow(rad1,2) - pow(diff0,2)
            factor = math.sqrt(factor) / rad1
            width = factor * delta_x
            height = factor * delta_y
            roi_pixel_matplot = self.__create_elipse(2 * width, 2 * height)
            point = super().mask_roi_in_slice(roi_pixel_matplot)
            for i in range(len(point)):
                point[i].append(number_slice)
            list_points.extend(point)
        return list_points

    def __create_elipse(self, width:float, height:float) -> matplotlib.patches.Ellipse: 
        """generate an ellipse patches from matplotlib

        Args:
            width (float): [widht of ellipse]
            height (float): [height of ellipse]

        Returns:
            [matplotlib.patches.Ellipse]: [Ellipse patche]
        """
        points_array = self.list_point_np  
        return matplotlib.patches.Ellipse(points_array[0], width, height, angle = 0)



    


import numpy as np 
from dicom_to_cnn.model.reader.Instance import Instance
from dicom_to_cnn.model.reader.Instance_RTSS import Instance_RTSS



class Instance_RTDose(Instance): 
    """A class to represent a RTDose Dicom File

    Args:
        Instance ([type]): [description]
    """

    def __init__(self, path_dose :str ):
        """constructor

        Args:
            path_dose (str): [path of the rt dose file ]
        """
        super().__init__(path_dose, load_image=True)

    def get_image_nparray(self) -> np.ndarray:
        """get 3d matrix of rt dose 

        Returns:
            [np.ndarray]: [return the 3D array from a RTDose file, shape (z,y,x) ]
        """
        return self.dicomData.pixel_array


    def wrap_DHVSequence_in_dict(self, rtss_file:str) -> dict: 
        """wrap in dictionnary DVH Sequence of each ROI 
        dict = { (roi) 1 : {'ReferencedROINumber' : value
                                'DVHType' : value
                                'DoseUnits' : value
                                'DVHDoseScaling' : value
                                'DVHVolumeUnits':  value
                                'DVHNumberOfBins' : value
                                'DVHMinimumDose' : value
                                'DVHMaximumDose' : value
                                'DVHMeanDose' : value
                                'DVHData' : value},
                (roi) 2 : {'ReferencedROINumber' : value
                                'DVHType' : value
                                'DoseUnits' : value
                                'DVHDoseScaling' : value
                                'DVHVolumeUnits':  value
                                'DVHNumberOfBins' : value
                                'DVHMinimumDose' : value
                                'DVHMaximumDose' : value
                                'DVHMeanDose' : value
                                'DVHData' : value}, ...}

        Returns:
            [dict]: [return a dictionnary, with, for each ROI, DVHSequence in the RTDose file]
        """

        instance_rtss_object = Instance_RTSS(rtss_file)
        dictionnary = {}
        for i in range(len(self.dicomData.DVHSequence)) : 
            referenced_roi = int(self.dicomData.DVHSequence[i].DVHReferencedROISequence[0].ReferencedROINumber)
            dataset = {}
            dataset['ReferencedROINumber'] = referenced_roi
            dataset['DVHType'] = self.dicomData.DVHSequence[i].DVHType
            dataset['DoseUnits'] = self.dicomData.DVHSequence[i].DoseUnits  
            dataset['DVHDoseScaling'] = self.dicomData.DVHSequence[i].DVHDoseScaling 
            dataset['DVHVolumeUnits'] = self.dicomData.DVHSequence[i].DVHVolumeUnits 
            dataset['DVHNumberOfBins'] = self.dicomData.DVHSequence[i].DVHNumberOfBins 
            dataset['DVHMinimumDose'] = self.dicomData.DVHSequence[i].DVHMinimumDose
            dataset['DVHMaximumDose'] = self.dicomData.DVHSequence[i].DVHMaximumDose
            dataset['DVHMeanDose'] = self.dicomData.DVHSequence[i].DVHMeanDose
            dataset['DVHData'] = self.dicomData.DVHSequence[i].DVHData 
            dataset['ReferencedROIName'] = instance_rtss_object.get_ROI_name(referenced_roi)
            dictionnary[referenced_roi] = dataset 

        return dictionnary



import numpy as np 
import SimpleITK as sitk 
from dicom_to_cnn.model.reader.Series import Series

def convert_array_to_img(numpy_array:np.ndarray, serie_object:Series) -> sitk.Image:
        """method to convert np.ndarray of shape (z,y,x,c) into pixel vector sitk Image of shape (x,y,z), associated with a PET Series

        Returns:
            [sitk.Image]: [return the corresponding sitk.Image from numpy_array]
        """
        instance_array = serie_object.get_instances_ordered()
        sitk_img = sitk.GetImageFromArray(numpy_array, isVector = True) # [z,y,x,c] tjrs de taille 4, si une seule roi=> dernier channel =1
        sitk_img = sitk.Cast(sitk_img, sitk.sitkVectorUInt8)
        original_pixel_spacing = instance_array[0].get_pixel_spacing()
        original_direction = instance_array[0].get_image_orientation()
        sitk_img.SetDirection( (float(original_direction[0]), float(original_direction[1]), float(original_direction[2]), 
                                    float(original_direction[3]), float(original_direction[4]), float(original_direction[5]), 
                                    0.0, 0.0, 1.0) )
        sitk_img.SetOrigin( instance_array[0].get_image_position())
        sitk_img.SetSpacing( (original_pixel_spacing[0], original_pixel_spacing[1], serie_object.get_z_spacing()) )
        return sitk_img

class DictSitk2Numpy(object):
    """
    Convert a dictionnary of SimpleITK image into Numpy ndarray
    """
    def __init__(self, keys:tuple):
        """constructor

        Args:
            keys (tuple): [tuple with key of sitk Image]
        """
        self.keys = keys

    def __call__(self, img_dict):
        
        for key in self.keys:
            img_dict[key] = sitk.GetArrayFromImage(img_dict.pop(key))
        
        return img_dict


from dicom_to_cnn.model.reader.Series import Series
from dicom_to_cnn.model.reader.Instance import Instance
import numpy as np 
import sys 



class Instance_RTSS(Instance):
    """A class to represent a RTSTRUCT Dicom file 
    """

    def __init__(self, path_rtss:str):
        """constructor 

        Args:
            path_rtss (str): [path of the rtss file]
        """
        super().__init__(path_rtss, load_image=True)


    def get_image_nparray(self):
        """Can't get image numpy array from a RTSTRUCT FILE 
        """
        sys.exit('Cannot get image numpy array from RTSTRUCT FILE')


    def get_list_all_SOP_Instance_UID_RTSS(self) -> list:
        """method to get all SOPInstanceUID referenced in RTSTRUCT File

        Returns:
            [list]: [list of all SOP Instance UID]
        """
        number_item = len(self.dicomData[0x30060010][0].RTReferencedStudySequence[0].RTReferencedSeriesSequence[0].ContourImageSequence)
        liste = []
        for i in range(number_item):
            liste.append(str(self.dicomData[0x30060010][0].RTReferencedStudySequence[0].RTReferencedSeriesSequence[0].ContourImageSequence[i].ReferencedSOPInstanceUID))
        return liste

    def get_number_of_roi(self) -> int:
        """method to count number of ROI in RTSTRUCT File

        Returns:
            [int]: [return the number of ROI in RTSS]
        """
        return len(self.dicomData.StructureSetROISequence)


    #info from ReferencedFrameOfReferenceSequence
    def get_number_of_referenced_series(self) -> int:
        """method to count the series referenced in ReferencedFrameOfReference Sequence

        Returns:
            [int]: [return the number of series referenced in ReferencedFrameOfReference Sequence]
        """
        return len(self.dicomData.ReferencedFrameOfReferenceSequence)


    def get_frame_of_reference_uid(self) -> list:
        """method to get FrameOfReferenceUID for each series referenced in ReferencedFrameOfReference Sequence.
            Usually, only one serie referenced.

        Returns:
            [list]: [list of FrameOfReferenceUID in ReferencedFrameOfReference Sequence]
        """
        number_serie = len(self.dicomData.ReferencedFrameOfReferenceSequence)
        liste = []
        for i in range(number_serie):
            liste.append(self.dicomData[0x30060010][i].FrameOfReferenceUID)
        return liste 


    def is_frame_of_reference_same(self, frame_of_ref_uid_serie:str) -> bool:
        """check if FrameOfReferenceUID in RTSS == FrameOfReferenceUID in associated serie 

        Args:
            frame_of_ref_uid_serie (str): [FrameOfReferenceUID in the associated serie]

        Returns:
            [bool]: [return True if UID in RTSS and UID in associated serie are equal, False instead. ]
        """
        frame_of_reference_UID = self.get_frame_of_reference_uid()
        for i in range(len(frame_of_reference_UID)):
            if frame_of_ref_uid_serie != frame_of_reference_UID[i] :
                return False 
        return True 


    def get_referenced_series_instance_uid(self) -> list:
        """method to get a list of SeriesInstanceUID for each series referenced in ReferencedFrameOfReference Sequence.
            Usually, only one serie referenced.
            Has to be the same SeriesInstanceUID in the associated serie. 

        Returns:
            [list]: [list of SeriesInstanceUID for each series referenced]
        """
        number_serie = len(self.dicomData.ReferencedFrameOfReferenceSequence)
        liste = []
        for i in range(number_serie):
            liste.append(self.dicomData[0x30060010][i].RTReferencedStudySequence[0].RTReferencedSeriesSequence[0].SeriesInstanceUID)
        return liste


    def get_referenced_study_SOP_instance_uid(self) -> list :
        """method to get a list of StudyInstanceUID for each series referenced in ReferencedFrameOfReference Sequence.
            Usually, only one serie referenced.
            Has to be the same StudyInstanceUID in the associated serie. 

        Returns:
            [list]: [list of StudyInstanceUID for each series referenced]
        """
        number_serie = len(self.dicomData.ReferencedFrameOfReferenceSequence)
        liste = []
        for i in range(number_serie):
            liste.append(self.dicomData[0x30060010][i].RTReferencedStudySequence[0].ReferencedSOPInstanceUID)
        return liste


    def get_referenced_SOP_class_UID(self) -> list :
        """method to get a list of SOPClassUID for each series referenced in ReferencedFrameOfReference Sequence.
            Usually, only one serie referenced.
            Has to be the same SOPCLassUID in the associated serie. 

        Returns:
            [list]: [list of SOPClassUID for each series referenced]
        """
        number_serie = len(self.dicomData.ReferencedFrameOfReferenceSequence)
        liste = []
        for i in range(number_serie):
            liste.append(self.dicomData[0x30060010][i].RTReferencedStudySequence[0].RTReferencedSeriesSequence[0].ContourImageSequence[0].ReferencedSOPClassUID)
        return liste 

 
    #info from StructureSetROISequence
    def get_ROI_name(self, number_roi:int) -> str:
        """method to get ROI name from StructureSetROI Sequence

        Args:
            number_roi (int): [ROI number; start at 1. ]

        Returns:
            [str]: [return ROI name]
        """
        return self.dicomData[0x30060020][number_roi - 1].ROIName 

    def get_ROI_volume(self, number_roi:int) -> str:
        """method to get ROI volume from StructureSetROI Sequence

        Args:
            number_roi (int): [ROI number; start at 1. ]

        Returns:
            [str]: [return ROI volume value]
        """
        return self.dicomData[0x30060020][number_roi - 1].ROIVolume 

    def get_ROI_generation_algorithm(self, number_roi:int) -> str :
        """method to get ROIGenerationAlgorithm value from StructureSetROI Sequence
        Args:
            number_roi (int): [ROI number; start at 1. ]

        Returns:
            [str]: [return ROIGenerationAlgorithme value]
        """
        return self.dicomData[0x30060020][number_roi - 1].ROIGenerationAlgorithm 


    #info from ROIContourSequence
    def get_roi_display_color(self, number_roi:int) -> str :
        """method to get ROI color representation from ROIContour Sequence

        Args:
            number_roi (int): [ROI number; start at 1. ]

        Returns:
            [str]: [return ROIDisplayColor value , format [r,g,b]]
        """
        return self.dicomData[0x30060039][number_roi - 1].ROIDisplayColor 


    def get_list_contour_SOP_Instance_UID(self, roi_number:int) -> list: 
        """method to get a list of every contour SOPINstanceUID  from ROIContour Sequence

        Args:
            number_roi (int): [ROI number; start at 1. ]

        Returns:
            [list]: [return list of every SOPInstanceUID where there is a contour]
        """
        number_item = len(self.dicomData[0x30060039][roi_number - 1].ContourSequence)
        liste = []
        for i in range(number_item):
            liste.append(str(self.dicomData[0x30060039][roi_number - 1].ContourSequence[i].ContourImageSequence[0].ReferencedSOPInstanceUID))
        return liste 

    
    def is_referenced_SOP_Instance_UID_in_all_SOP_Instance(self) -> bool :
        """check if every  SOPInstanceUID contour is referenced in the SOPINstanceUID of the associated serie

        Raises:
            Exception: [Exception if one of the SOPInstanceUID contour is not referenced in all SOPInstanceUID of the associated serie]

        Returns:
            [bool]: [True if everything referenced, raise Exception instead]
        """
        all_sop = str(self.get_list_all_SOP_Instance_UID_RTSS)
        referenced_sop = str(self.get_list_contour_SOP_Instance_UID)
        for uid in referenced_sop : 
            if uid not in all_sop : 
                raise Exception("SOP Instance UID not in the serie")
        return True 


    def get_number_of_contour_points(self, roi_number:int) -> list:
        """method to get a list of every NumberOfContourPoint in each contours  from ROIContour Sequence

        Args:
            number_roi (int): [ROI number; start at 1. ]

        Returns:
            [list]: [return list of number of contour points]
        """
        number_item = len(self.dicomData[0x30060039][roi_number - 1].ContourSequence)
        liste = []
        for i in range(number_item):
            liste.append(self.dicomData[0x30060039][roi_number - 1].ContourSequence[i].NumberOfContourPoints)
        return liste

    def get_list_contour_geometric_type(self, roi_number:int) -> list :
        """method to get a list of every ContourGeomtricType value of each contour  from ROIContour Sequence

        Args:
            number_roi (int): [ROI number; start at 1. ]

        Returns:
            [list]: [return a list with every ContourGeometricType value for each contour]
        """
        number_item = len(self.dicomData[0x30060039][roi_number - 1].ContourSequence)
        liste = []
        for i in range(number_item):
            liste.append(self.dicomData[0x30060039][roi_number - 1].ContourSequence[i].ContourGeometricType)
        return liste

    def is_closed_planar(self, roi_number:int) -> bool :
        """check if every contour is CLOSED_PLANAR  from ROIContour Sequence

        Args:
            roi_number (int): [ROI number, start a 1]

        Returns:
            [bool]: [return True if every value is CLOSED_PLANAR, False instead]
        """
        geometric_type = self.get_list_contour_geometric_type(roi_number)
        for type in geometric_type :
            if type != "CLOSED_PLANAR" : 
                return False
        return True 

    def get_contour_data(self, roi_number:int) -> list :
        """method to get contour data of contours  from ROIContour Sequence

        Args:
            number_roi (int): [ROI number; start at 1. ]

        Returns:
            [list]: [return list of contour data for each contour [[contour1 : x, y, z, x, y ,z, ...], [contour2 : ...], [...]] ]
        """
        number_item = len(self.dicomData[0x30060039][roi_number - 1].ContourSequence)
        liste = []
        for i in range(number_item):
            liste.append(self.dicomData[0x30060039][roi_number - 1].ContourSequence[i].ContourData)
        return liste 



from dicom_to_cnn.model.reader.Series import Series
import numpy as np
import SimpleITK as sitk 

class SeriesCT(Series):
    """Get Series CT Nifti in 16 Bits

    Arguments:
        Series {String} -- Series Location Path
    """

    def __init__(self,path:str):
        """constructor

        Args:
            path (str): [path folder of CT serie ]
        """
        super().__init__(path)

    def get_numpy_array(self) -> np.ndarray:
        numpy_array = super().get_numpy_array()
        return numpy_array.astype(np.int16)

    def export_nifti(self, file_path:str):
        """method to export/save ndarray of series to nifti format

        Args:
            file_path (str): [directory+filename of the nifti]
        """
        sitk_img = sitk.GetImageFromArray( np.transpose(self.get_numpy_array(), (2,0,1) ))
        sitk_img = sitk.Cast(sitk_img, sitk.sitkInt16)
        original_pixel_spacing = self.instance_array[0].get_pixel_spacing()        
        original_direction = self.instance_array[0].get_image_orientation()
        sitk_img.SetDirection( (float(original_direction[0]), float(original_direction[1]), float(original_direction[2]), 
                                    float(original_direction[3]), float(original_direction[4]), float(original_direction[5]), 
                                    0.0, 0.0, 1.0) )
        sitk_img.SetOrigin( self.instance_array[0].get_image_position() )
        sitk_img.SetSpacing((original_pixel_spacing[0], original_pixel_spacing[1], self.get_z_spacing()) )
        sitk.WriteImage(sitk_img, file_path)


import SimpleITK as sitk 
import numpy as np 

class Nifti : 
    """a class to read a nifti image, resample it 
    """

    def __init__(self, nifti_img_path:str):
        """constructor

        Args:
            nifti_img_path (str): [path of a nifti image with shape (x,y,z)]
        """
        self.nifti_img_path = nifti_img_path
        self.nifti_img = sitk.ReadImage(self.nifti_img_path) #shape (x,y,z)

    def resample(self, shape_matrix:tuple = (256, 256, 1024), shape_physic=(700, 700, 2000)) -> np.ndarray:
        """function to resample sitk image of shape (x,y,z), put it in a bigger and empty np.ndarray

        Args:
            shape_matrix (tuple) : [size of the new matrix/sitk image (number of pixel), shape (x,y,z)].
            shape_physic (tuple) : [physical size of the new matrix/sitk image, (in mm), shape (x,y,z)].
        Returns:
            [np.ndarray]: [return corresponding resampled np.ndarray]
        """
        spacing = self.nifti_img.GetSpacing()
        origin = self.nifti_img.GetOrigin()
        direction = self.nifti_img.GetDirection()
        size = self.nifti_img.GetSize()
        #target spacing, and size
        spacing_x = shape_physic[0]/shape_matrix[0] #mm
        spacing_y = shape_physic[1]/shape_matrix[1] #mm 
        spacing_z = shape_physic[2]/shape_matrix[2] #mm

        true_x = size[0] * spacing[0] #mm
        true_y = size[1] * spacing[1] #mm 
        true_z = size[2] * spacing[2] #mm

        new_size_x = int((true_x * shape_matrix[0]) / shape_physic[0]) #pixel
        new_size_y = int((true_y * shape_matrix[1]) / shape_physic[1]) #pixel
        new_size_z = int((true_z * shape_matrix[2]) / shape_physic[2]) #pixel

        #applied transformation
        transformation = sitk.ResampleImageFilter()
        transformation.SetOutputDirection(direction)
        transformation.SetOutputOrigin(origin)
        transformation.SetSize((new_size_x, new_size_y, new_size_z))
        transformation.SetOutputSpacing((spacing_x, spacing_y, spacing_z))
        transformation.SetInterpolator(sitk.sitkLinear)
        new_img = transformation.Execute(self.nifti_img) 
        result = sitk.GetArrayFromImage(new_img) #[z,y,x]
        center = [int(shape_matrix[2]/2), int(shape_matrix[1]/2),  int(shape_matrix[1]/2)]
        z = int(result.shape[0]/2)
        y = int(result.shape[1]/2)
        x = int(result.shape[2]/2)
        sommet_x = center[2] - x 
        sommet_y = center[1] - y 
        sommet_z = center[0] - z
        new_array = np.zeros((shape_matrix[2], shape_matrix[1], shape_matrix[0]))
        if result.shape[1] != shape_matrix[1] : 
            new_array[sommet_z:sommet_z+result.shape[0], sommet_y:sommet_y + result.shape[1], sommet_x:sommet_x + result.shape[2]] = result
        else : 
            new_array[sommet_z:sommet_z+result.shape[0],0:shape_matrix[1], 0:shape_matrix[0]] = result
        return new_array





from numpy import random 
import os 
from dicom_to_cnn.tools.cleaning_dicom.folders import *


def generate_dict(number_of_roi:int, name:str, serie_description:str, body_part:str, roi_name:dict, interpreted_type:dict=None) -> dict:
    """a function to generate a dict with informations to generate DICOM SEG or DICOM RTSTRUCT file
    ! no more than 16 characters for each informations ! 

    Args:
        number_of_roi ([int]): [Total number of ROI in dicom file ]
        mode ([str]): ['rtstruct' or 'dicomseg']

    Returns:
        [dict]: [return dict with informations about ROI in it ]
    """

    results = {}
    results["ContentCreatorName"] = "dicom_to_cnn"
    random_number = random.randint(0,1e3)
    
    results["ClinicalTrialSeriesID"] = name #name of json

    results["ClinicalTrialTimePointID"] = "1"

    results["SeriesDescription"] = serie_description
    
    results["SeriesNumber"] = str(random_number)
    results["InstanceNumber"] = str(number_of_roi)

    results["BodyPartExamined"] = body_part

    results["segmentAttributes"] = []
    subliste = []
    for i in range(number_of_roi):
        subdict = {}
        subdict["labelID"]= i+1
        subdict["SegmentDescription"] = roi_name[i+1]

        subdict["SegmentAlgorithmType"] = "SEMIAUTOMATIC"
        subdict["SegmentAlgorithmName"] =  "DeepOncology"
        if interpreted_type is not None : 
            subdict['RTROIInterpretedType'] = interpreted_type[i+1]
        
        #subdict["SegmentedPropertyCategoryCodeSequence"] 
        subsubdict = {}
        subsubdict["CodeValue"] = "49755003"
        subsubdict["CodingSchemeDesignator"] = "SCT"
        subsubdict["CodeMeaning"] = "Morphologically Altered Structure"
        subdict["SegmentedPropertyCategoryCodeSequence"] = subsubdict

        #subdict["SegmentedPropertyTypeCodeSequence"] = {}
        subsubdict = {}
        subsubdict["CodeValue"] = "108369006"
        subsubdict["CodingSchemeDesignator"] = "SCT"
        subsubdict["CodeMeaning"] = "Neoplasm"
        subdict["SegmentedPropertyTypeCodeSequence"] = subsubdict
        subliste.append(subdict)

    results["segmentAttributes"].append(subliste)

    results["ContentLabel"] =  "SEGMENTATION"
    results["ContentDescription"] = "Image segmentation"
    results["ClinicalTrialCoordinatingCenterName"] = "dicom_to_cnn"

    return results

def save_dict_as_json(results:dict, directory:str) -> str :
    """a function to save the dict generated by 'generate_dict' function above

    Args:
        results ([dict]): [Dictionnary generated by 'generate_dict' function ]
        directory ([str]): [Name of directory to save the dict as json file ]

    Returns:
        [str]: [return the path name of the json file]
    """
    filename = results["ClinicalTrialSeriesID"]
    write_json_file(directory, filename, results)
    return os.path.join(directory, filename+'.json')



from dicom_to_cnn.model.petctviewer.Roi import Roi

class RoiNifti(Roi):
    """Derivated Class for automatic Nifti ROI of PetCtViewer.org

    Returns:
        [RoiNifti] -- Nifti ROI
    """

    def __init__(self, roi_number:int, list_point:list, volume_dimension:tuple):
        """constructor

        Args:
            roi_number (int): [roi number]
            list_point (list): [list of [x,y,z] coordonates ]
            volume_dimension (tuple): [(shape x, shape y, shape z)]
        """
        super().__init__(1, 0, 0, roi_number, 0, list_point, volume_dimension)
        self.list_points = self.calculateMaskPoint()

    def calculateMaskPoint(self) -> list:
        """Return [x,y,z] coordonates/voxel of nifti ROI (already in CSV file)

        Returns:
            [list]: [list of [x,y,z] coordonates of ROI ]
        """
        pixel_array = self.list_point
        list_points = []
        for points in pixel_array :
            if len(points) != 0 : 
                list_points.append(points)
        return list_points


import numpy as np

def threshold_matrix(mask_array:np.ndarray, pet_array:np.ndarray, threshold:float) -> np.ndarray:
    """function to threshold a ndarray mask

    Args:
        mask_array (np.ndarray): [(z,x,y,c)]
        pet_array (np.ndarray): [(z,x,y)]
        threshold (float): [0.41, 2.5, 4.0 or other]

    Returns:
        [ndarray]: [threshold mask]
    """
   
    mask_array = mask_array.astype('int8')
    if threshold < 1 : 
        if len(mask_array.shape) != 3 :  #mask 4D 
            number_of_roi = mask_array.shape[3]
            liste = []
            for i in range(number_of_roi):
                new_mask = np.zeros((mask_array.shape[0],mask_array.shape[1], mask_array.shape[2] )).astype('uint8')
                suv_values = []
                x,y,z = np.where(mask_array[:,:,:,i] != 0)
                if len(x) != 0 : 
                    suv_values = pet_array[x,y,z].tolist()
                    seuil = np.max(suv_values) * threshold
                    new_mask[np.where((pet_array > seuil) & (mask_array[:,:,:,i] > 0))] = 1
                    liste.append(new_mask.astype('uint8'))
                else : liste.append(new_mask.astype('uint8'))
            return np.stack(liste, axis = 3).astype('uint8')   

        else : 
            maxi = np.max(mask_array)
            if maxi == 1.0 : 
                new_mask = np.zeros((mask_array.shape[0],mask_array.shape[1], mask_array.shape[2] )).astype('uint8')
                x,y,z = np.where(mask_array != 0)
                suv_values = pet_array[x,y,z].tolist()
                seuil = np.max(suv_values) * threshold
                new_mask[np.where((pet_array > seuil) & (mask_array > 0))] = 1
                return new_mask.astype('uint8')
            else :  
                number_of_roi = maxi
                new_mask = np.zeros((mask_array.shape[0],mask_array.shape[1], mask_array.shape[2] )).astype('uint8')
                for i in range(number_of_roi):
                    suv_values = []
                    x,y,z = np.where(mask_array == i+1)
                    suv_values = pet_array[x,y,z].tolist()
                    seuil = np.max(suv_values) * threshold

                    new_mask[np.where((pet_array > seuil) & (mask_array == i+1))] = i+1
                return new_mask.astype('uint8')

    else : 
        if len(mask_array.shape) != 3 :  #mask 4D 
            number_of_roi = mask_array.shape[3]
            liste = []
            for i in range(number_of_roi):
                new_mask = np.zeros((mask_array.shape[0],mask_array.shape[1], mask_array.shape[2] )).astype('uint8')
                seuil = threshold
                new_mask[np.where((pet_array > seuil) & (mask_array[:,:,:,i] > 0))] = 1
                liste.append(new_mask)
            return np.stack(liste, axis = 3)    

        else : 
            maxi = np.max(mask_array)
            if maxi == 1.0 : 
                new_mask = np.zeros((mask_array.shape[0],mask_array.shape[1], mask_array.shape[2] )).astype('uint8')
                seuil = threshold
                new_mask[np.where((pet_array > seuil) & (mask_array > 0))] = 1
                return new_mask.astype('uint8')
            else :  
                number_of_roi = maxi
                new_mask = np.zeros((mask_array.shape[0],mask_array.shape[1], mask_array.shape[2] )).astype('uint8')
                seuil = threshold
                for i in range(number_of_roi):
                    new_mask[np.where((pet_array > seuil) & (mask_array == i+1))] = i+1
                return new_mask.astype('uint8')








import pydicom_seg 
import pydicom 
import numpy as np 
import os
from dicom_to_cnn.tools.export_segmentation.generate_dict import *
from dicom_to_cnn.model.segmentation.Abstract_Writer import Abstract_Writer
import SimpleITK as sitk 


class DICOMSEG_Writer(Abstract_Writer):
    """a class to write a DICOMSEG file
    """

    def __init__(self, mask_img:sitk.Image, serie_path:str):
        """constructor

        Args:
            img_mask ([sitk.Image]): [sitk img of segmentation, [x,y,z]]
            serie_path ([str]): [Serie path related to DICOMSEG file]
        """
        self.mask_img = mask_img
        self.serie_path = serie_path

    def setDictName(self, name:str) -> str : 
        """set the dict name 

        Args:
            name (str): [dict name]

        """
        self.dict_name = name
        
    def setBodyPartExaminated(self, name:str) -> str : 
        """set the body part examinated

        Args:
            name (str): [name of the body part]

        """
        if len(name) > 16 : self.body_part =  name[0:16]
        
        else : self.body_part = name

    def setSeriesDescription(self, description:str) -> str : 
        """set the study description of the exam

        Args:
            description (str): [study description]

        """
        if len(description) > 16 : self.serie_description = description[0:16]
        else : self.serie_description =  description 

    def setRoiName(self, dictionnary:dict) -> dict : 
        """set a dict for ROI Name, for each ROI

        Args:
            dictionnary (dict): [{1 : 'name roi 1', 
                                  2:'name roi 2', 
                                 etc }]

        """
        self.roi_name_dict = dictionnary

    def setAutoRoiName(self) -> dict : 
        """set an auto dict with ROI Name value 

            generated dict  {1 : 'ROI 1', 
                            2 : 'ROI 2', 
                             etc }]
        """
        mask_array = sitk.GetArrayFromImage(self.mask_img)
        number_of_roi = int(np.max(mask_array))
        dictionnary = {}
        for i in range(1, number_of_roi+1):
            dictionnary[i] = str("ROI {}".format(i+1))
        self.roi_name_dict =  dictionnary


    def __generate_dict_json(self, directory_path:str) -> str:
        """method to generate dict with metainfo for DICOM SEG file and save it as json file

        Args:
            directory_path (str): [directory's path where to save the json file with metainfo]

        Returns:
            [str]: [return the path of the json generated]
        """
        number_of_roi = int(np.max(sitk.GetArrayFromImage(self.mask_img)))
        results = generate_dict(number_of_roi, self.dict_name, self.serie_description, self.body_part, self.roi_name_dict, interpreted_type=None)
        self.results = results
        json_path = save_dict_as_json(results, directory_path)
        return json_path 

    def dicom_seg_writer(self, directory_path:str) -> pydicom.FileDataset:
        """method to write a DICOMSEG file from a segmentation

        Args:
            directory_path (str): [directory's path where to save the json file with metainfo]

        Returns:
            [pydicom.FileDataset]: [return the new DICOMSEG]
        """
        json_path = self.__generate_dict_json(directory_path)

        segmentation = self.mask_img
        dicom_file = os.listdir(self.serie_path)
        dicom_series_paths = [os.path.join(self.serie_path, x) for x in dicom_file]
        source_images = [
            pydicom.dcmread(x, stop_before_pixels=True)
            for x in dicom_series_paths]

        template = pydicom_seg.template.from_dcmqi_metainfo(json_path)

        writer = pydicom_seg.MultiClassWriter(
                template=template,
                inplane_cropping=False,  # Crop image slices to the minimum bounding box on
                                        # x and y axes. Maybe not supported by other frameworks.
                skip_empty_slices=True,  # Don't encode slices with only zeros
                skip_missing_segment=False,  # If a segment definition is missing in the
                                            # template, then raise an error instead of
                                            # skipping it.
            )

        dcm = writer.write(segmentation, source_images)
        os.remove(json_path)
        return dcm


    def save_file(self, filename:str, directory_path:str) -> None :
        """method to save the new DICOMSEG file

        Args:
            filename (str): [name of the new DICOMSEG file]
            directory_path (str): [directory's path where to save the new DICOMSEG file]

        """
        dcm = self.dicom_seg_writer(directory_path)
        dcm.save_as(os.path.join(directory_path, filename))


from dicom_to_cnn.model.petctviewer.CsvReader import CsvReader
import numpy as np 

"""functions to get informations about ROI volume in CSV file generated by petctviewer.org
"""
def get_total_volume(csv_path:str) -> float:
    """get total volume of all ROI in CSV file

    Args:
        csv_path (str): [path of the CSV file]

    Returns:
        [float]: [total volume of tumor]
    """
    csv_objet = CsvReader(csv_path)
    total_volume = float(csv_objet.get_sum_row()[3])
    return total_volume


def keep_csv_higher_volume(liste_csv_path:list) -> list:
    """function to

    Args:
        liste_csv_path ([list]): [list of CSV file path]

    Returns:
        ([list]) : [return the CSV file path with the higher volume]
    """
    volume = []
    for csv_file in liste_csv_path : 
        volume.append(get_total_volume(csv_file))
    index_maxi = volume.index(np.max(volume))
    if type(index_maxi) == list: 
        index_maxi = index_maxi[0]
    index = np.arange(len(volume)).tolist()
    remove = []
    for i in index:
        if i != index_maxi : 
            remove.append(liste_csv_path[i])
    for r in remove : 
        liste_csv_path.remove(r)
    return liste_csv_path


def get_sum_rois_volume(list_results_row:list) -> float:
    """sum all ROIs volume

    Args:
        liste ([list]): [List of every ROI results row from CSV file]

    Returns:
        [float]: [Return the sum of ROIs volume]
    """
    volume = float(0)
    for i in range(len(list_results_row)):
        volume += float(list_results_row[i][3])

    return volume


def get_difference_volume(csv_path:str) -> tuple :
    """Calcul the difference between volume and sum of every ROI volume

    Args:
        csv_path ([str]): [CSV file path]

    Returns:
        [tuple]: [Return total_volume, sum of every ROIs volume, and difference between the 2]
    """
    csv_objet = CsvReader(csv_path)
    total_volume = float(csv_objet.get_sum_row()[3])

    liste = csv_objet.get_rois_results()

    sum_volume = get_sum_rois_volume(liste)
    return total_volume, sum_volume, abs(sum_volume - total_volume)





import numpy as np 
import SimpleITK as sitk 

def label_stat_results(labelled_img:sitk.Image, pet_img:sitk.Image) -> dict :
    """a function to gather stats about each ROI

    Args:
        labelled_img (sitk.Image): [3D segmentation sitk.Image with label]
        pet_img (sitk.Image): [3D pet sitk.Image]

    Returns:
        [type]: [description]
    """
    results = {} 
    stats = sitk.LabelIntensityStatisticsImageFilter()
    stats.Execute(labelled_img, pet_img)
    pet_spacing = pet_img.GetSpacing()
    number_of_label = stats.GetNumberOfLabels()
    results['number_of_label'] =  number_of_label
    volume = 0
    for i in range(1, number_of_label + 1) :
        subresult = {}
        subresult['max'] = stats.GetMaximum(i)
        subresult['mean'] = stats.GetMean(i)
        subresult['median'] = stats.GetMedian(i)
        subresult['variance'] = stats.GetVariance(i)
        subresult['sd'] = stats.GetStandardDeviation(i)
        subresult['number_of_pixel'] = stats.GetNumberOfPixels(i)
        volume_voxel = pet_spacing[0] * pet_spacing[1] * pet_spacing[2] * 10**(-3) #ml
        subresult['volume'] = stats.GetNumberOfPixels(i) * volume_voxel
        subresult['centroid'] = stats.GetCentroid(i)
        volume += stats.GetNumberOfPixels(i) * volume_voxel
        results[i] = subresult
    results['total_vol'] = volume
    return results




import setuptools 

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="dicom_to_cnn",                     # This is the name of the package
    version="0.76",                        # The initial release version
    author="",                     # Full name of the author
    description="Python Library to handle Input / Output conversion in Dicom <=> Convolutional Neural Network ",
    long_description=long_description,      # Long description read from the the readme file
    long_description_content_type="text/markdown",
    url = 'https://github.com/wendyrvllr/Dicom-To-CNN',   # Provide either the link to your github or to your website
    download_url = 'https://github.com/wendyrvllr/Dicom-To-CNN/archive/refs/tags/76.tar.gz',
    packages=setuptools.find_packages(),    # List of all python modules to be installed
    classifiers=[
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],                                      # Information to filter the project on PyPi website
    python_requires='>=3.7',                # Minimum version requirement of the package
    install_requires=['fpdf==1.7.2',
                    'imageio==2.9.0',
                    'opencv_python==4.5.1.48',
                    'matplotlib==3.4.2',
                    'numpy>=1.19',
                    'pandas==1.2.3',
                    'Pillow==8.2.0',
                    'plotly_express==0.4.1',
                    'pydicom_seg==0.2.3',
                    'pydicom==2.1.2',
                    'pyradiomics==3.0.1',
                    'scikit-image==0.18.1',
                    'scipy==1.6.3',
                    'SimpleITK==2.0.2']                     # Install other dependencies if any
)


import pydicom_seg
import pydicom 
import numpy as np 
import SimpleITK as sitk 

class MaskBuilder_DICOMSEG: 
    """a class to read a DICOMSEG file format and generate np.ndarray mask 
    """

    def __init__(self, path_dicomseg_file:str):
        """constructor

        Args:
            path_dicomseg_file (str): [abs path to the dicom_seg file]
        """

        self.path_dicomseg_file = path_dicomseg_file
        self.reader = self.__reader()

    def __reader(self):
        dcm = pydicom.dcmread(self.path_dicomseg_file)
        reader = pydicom_seg.MultiClassReader()
        result = reader.read(dcm)
        return result

    def get_numpy_array(self) -> np.ndarray:
        """return the np.ndarray of the segmentation 

        Returns:
            [np.ndarray]: [return the segmentation mask np.ndarray of shape [z,y,x]]
        """
        return self.reader.data 

    def get_sitk_image(self) -> sitk.Image:
        """return the sitk.Image of the segmentation 

        Returns:
            [np.ndarray]: [return the segmentation mask sitk.Image of shape [x,y,z]]
        """
        return self.reader.image 


import os
import json

def get_series_path(path:str) -> list :
    """Go through all the folder to find every series path

    Arguments:
        path {[string]} -- [Absolute path where the repertory is located]

    Returns:
        [list] -- [Path's list of every series]
    """

    seriesPath = []
    for (path, dirs, files) in os.walk(path): 
        if not (dirs) :
            seriesPath.append(path)
    return seriesPath


def write_json_file(path:str, file_name:str, content:list, extension:str = 'json'):
    """write and save a json file

    Args:
        path (str): [directory's path where to save json file]
        file_name (str): [name of the json file]
        content (list): [what to write in the json file]
        extension (str, optional): [description]. Defaults to 'json'.
    """
    wrinting_file = open(os.path.join(path, file_name+'.'+extension), 'w')
    wrinting_file.write(json.dumps(content))
    wrinting_file.close()

def remove_bi_file(path:str):
    """remove bi file of a dicom serie

    Args:
        path (str): [directory's path of the serie]
    """
    file_names = os.listdir(path)
    if('graphic.brownFat.gr2' in file_names):
        os.remove( os.path.join(path,"graphic.brownFat.gr2") )

def remove_index_ini(path:str):
    """remove index ini of a dicom serie

    Args:
        path (str): [directory's path of the serie]
    """
    file_names = os.listdir(path)
    if ('index.ini' in file_names):
        os.remove(os.path.join(path, 'index.ini'))

def remove_empty_folders(path:str):
    """remove serie with no dicom file in it 

    Args:
        path (str): [directory's path of a dicom serie]

    """
    list_instances = os.listdir(path)
    if len(list_instances) == 0 : 
        os.remove(path)



import pydicom 

def create_ContourImageSequence(ReferencedSOPClassUID:str, list_all_SOPInstanceUID:list) -> pydicom.Sequence:
    """function to generate ContourImageSequence from ReferencedFrameOfReferenceSequence

    Args:
        ReferencedSOPClassUID (str): [Referenced SOPCLAss UID value from associated serie ]
        list_all_SOPInstanceUID (list): [list of every SOPINstanceUID of every slice in associated dicom serie]

    Returns:
        [pydicom.Sequence]: [return ContourImageSequence]
    """
    ContourImageSequence = pydicom.sequence.Sequence()
    for SOPInstanceUID in list_all_SOPInstanceUID : 

        dataset = pydicom.dataset.Dataset()
        dataset.ReferencedSOPClassUID = ReferencedSOPClassUID
        dataset.ReferencedSOPInstanceUID = SOPInstanceUID
        ContourImageSequence.append(dataset)
    return ContourImageSequence 

def create_RTReferencedSeriesSequence(ReferencedSOPClassUID:str, list_all_SOPInstanceUID:list, SeriesInstanceUID:str) -> pydicom.Sequence:
    """function to generate RTReferencedSeriesSequence from ReferencedFrameOfReferenceSequence

    Args:
        ReferencedSOPClassUID (str): [Referenced SOPCLAss UID value from associated value]
        list_all_SOPInstanceUID (list): [list of every SOPINstanceUID of every slice in associated dicom serie]
        SeriesInstanceUID (str): [SeriesInstanceUID value]

    Returns:
        [pydicom.Sequence]: [return RTReferencedSeriesSequence]
    """
    RTReferencedSeriesSequence = pydicom.sequence.Sequence()
         
    dataset = pydicom.dataset.Dataset()
    dataset.SeriesInstanceUID = SeriesInstanceUID
    dataset.ContourImageSequence = create_ContourImageSequence(ReferencedSOPClassUID, list_all_SOPInstanceUID)
    RTReferencedSeriesSequence.append(dataset)
    return RTReferencedSeriesSequence

def create_RTReferencedStudySequence(ReferencedSOPClassUID:str, list_all_SOPInstanceUID:list, SeriesInstanceUID:str, StudyInstanceUID:str) -> pydicom.Sequence:
    """function to generate RTReferencedStudySequence from ReferencedFrameOfReferenceSequence

    Args:
        ReferencedSOPClassUID (str): [Referenced SOPCLAss UID value from associated serie]
        list_all_SOPInstanceUID (list): [list of every SOPINstanceUID of every slice in associated dicom serie]
        SeriesInstanceUID (str): [SeriesInstanceUID value from associated serie]
        StudyInstanceUID (str): [StudyInstanceUID value from associated serie]

    Returns:
        [pydicom.Sequence]: [return RTReferencedStudySequence ]
    """
    RTReferencedStudySequence = pydicom.sequence.Sequence()
    dataset = pydicom.dataset.Dataset()
    dataset.ReferencedSOPInstanceUID = StudyInstanceUID
    dataset.ReferencedSOPClassUID = ReferencedSOPClassUID
    dataset.RTReferencedSeriesSequence = create_RTReferencedSeriesSequence(ReferencedSOPClassUID, list_all_SOPInstanceUID, SeriesInstanceUID)
    RTReferencedStudySequence.append(dataset)
    return RTReferencedStudySequence

def create_ReferencedFrameOfReferenceSequence(FrameOfReferenceUID:str, ReferencedSOPClassUID:str, list_all_SOPInstanceUID:list, SeriesInstanceUID:str, StudyInstanceUID:str) -> pydicom.Sequence: 
    """function to generate ReferencedFrameOfReferenceSequence

    Args:
        FrameOfReferenceUID (str): [FrameOfReferenceUID value from associated serie]
        ReferencedSOPClassUID (str): [Referenced SOPCLAss UID value from associated serie]
        list_all_SOPInstanceUID (list): [list of every SOPINstanceUID of every slice in associated dicom serie]
        SeriesInstanceUID (str): [SeriesInstanceUID value from associated serie]
        StudyInstanceUID (str): [StudyInstanceUID value from associated serie]


    Returns:
        [pydicom.Sequence]: [return ReferencedFrameOfReferenceSequence from RTSTRUCT file]
    """
    ReferencedFrameOfReferenceSequence = pydicom.sequence.Sequence()
    dataset = pydicom.dataset.Dataset()
    dataset.FrameOfReferenceUID = FrameOfReferenceUID 
    dataset.RTReferencedStudySequence = create_RTReferencedStudySequence(ReferencedSOPClassUID, list_all_SOPInstanceUID, SeriesInstanceUID, StudyInstanceUID)
    ReferencedFrameOfReferenceSequence.append(dataset)
    return ReferencedFrameOfReferenceSequence



import SimpleITK as sitk 
import os
from dicom_to_cnn.model.reader.SeriesPT import SeriesPT
from dicom_to_cnn.model.segmentation.Abstract_Writer import Abstract_Writer

class Nifti_Writer(Abstract_Writer):
    """a class to convert a np.ndarray to a sitk.Image  
    """

    def __init__(self, mask_img:sitk.Image):
        """constructor

        Args:
            img_mask (sitk.Image) : [sitk.Image of shape (x,y,z)]
        """
        self.mask_img = mask_img

    def save_file(self, filename:str, directory_path:str) -> None :
        """method to save the new nifti file

        Args:
            filename (str): [name of the new nifti file]
            directory_path (str): [directory's path where to save the new nifti file]
        
        """
        
        sitk.WriteImage(self.mask_img, os.path.join(directory_path, filename))